import numpy as np
from os.path import join,exists
from os import remove
from numpy import array_equal
import h5py

from emdfile import _TESTPATH
from emdfile import save,read,set_author
from emdfile.read import _is_EMD_file,_get_EMD_rootgroups
from emdfile.classes import (
    Node,
    Root,
    Metadata,
    Array,
    PointList,
    PointListArray
)

# Set paths
dirpath = _TESTPATH
path_h5 = join(dirpath,"test.h5")





class TestHeader:

    ## Setup and teardown

    @classmethod
    def setup_class(cls):
        cls._clear_files(cls)
        cls._make_data(cls)

    @classmethod
    def teardown_class(cls):
        pass

    def setup_method(self, method):
        pass

    def teardown_method(self, method):
        self._clear_files()

    def _make_data(self):
        """ Make
            - an array
            - an array with some 'slicelabels'
            #- a pointlist
            #- a pointlistarray
        """
        # arrays
        self.array = Array(
            data = np.arange(np.prod((2,3))).reshape((2,3))
        )



    def _clear_files(self):
        """
        Delete h5 files which this test suite wrote
        """
        paths = [
            path_h5
        ]
        for p in paths:
            if exists(p):
                remove(p)





    ## Tests

    def test_header(self):
        """ Save a file, and check its header tags
        """
        # save
        save(path_h5,self.array)

        # check tags
        with h5py.File(path_h5,'r') as f:
            for key in (
                'emd_group_type',
                'version_major',
                'version_minor',
                'authoring_program',
                'authoring_user',
                'UUID'
            ):
                assert( key in f.attrs )

            assert( f.attrs['emd_group_type'] == 'file' )
            assert( f.attrs['version_major'] == 1 )
            assert( f.attrs['version_minor'] == 0 )
            assert( f.attrs['authoring_program'] == 'emdfile' )
            assert( f.attrs['authoring_user'] == '' )
            uuid = f.attrs['UUID']
            assert(isinstance(uuid,str))



    def test_header2(self):
        """ Save a file after modifying the authoring user, then
        check its header tags
        """
        set_author('emdfile_test_suite')


        # save
        save(path_h5,self.array)

        # check tags
        with h5py.File(path_h5,'r') as f:
            for key in (
                'emd_group_type',
                'version_major',
                'version_minor',
                'authoring_program',
                'authoring_user',
                'UUID'
            ):
                assert( key in f.attrs )

            assert( f.attrs['emd_group_type'] == 'file' )
            assert( f.attrs['version_major'] == 1 )
            assert( f.attrs['version_minor'] == 0 )
            assert( f.attrs['authoring_program'] == 'emdfile' )
            assert( f.attrs['authoring_user'] == 'emdfile_test_suite' )
            uuid = f.attrs['UUID']
            assert(isinstance(uuid,str))






# Defines the Custom class, which contains any number of EMD base
# classes as attributes, i.e. they are constructed by composition,
# rather than inheritance, of the other classes.  Custom classes are
# themselves EMD base classes and can therefore be nested in other custom
# classes themselves.

import h5py

from emdfile.classes.tree import Node,Root
from emdfile.classes.utils import _get_class

class Custom(Node):
    """
    """
    _emd_group_type = 'custom'

    def __init__(self,name='custom'):
        Node.__init__(self,name=name)



    # read

    @classmethod
    def _get_constructor_args(cls,group):
        """
        Takes a group, and returns a dictionary of args/values
        to pass to the class constructor.

        This function needs to be written for any child classes;
        subclassing but failing to overwrite this method will
        throw a runtime error. You'll probably want to include
        {'name' : basename(group.name)}. To retrieve the data stored in
        a Python class instance attribute, use self._get_emd_attr_data.
        To populate subsequent to instance construction, add a
        _populate_instance method.
        """
        raise Exception(f"Custom class {cls} needs a `_get_constructor_args` method!")


    def _get_emd_attr_data(self,group):
        """ Loops through h5 groups under group, finding and returning
        a {name:instance} dictionary of {attribute name : class instance}
        pairs from all 'custom_' subgroups
        """
        groups = [g for g in group.keys() if isinstance(group[g],h5py.Group)]
        groups = [g for g in groups if 'emd_group_type' in group[g].attrs.keys()]
        groups = [g for g in groups if group[g].attrs['emd_group_type'][:7]=='custom_']
        dic = {}
        for g in groups:
            name,grp = g,group[g]
            __class__ = _get_class(grp)
            data = __class__.from_h5(grp)
            dic[name] = data
        return dic



    # write
    def to_h5(self,group):
        """
        Constructs an h5 group, adds metadata, and adds all attributes
        which point to EMD nodes.

        Accepts:
            group (h5py Group)

        Returns:
            (h5py Group) the new node's Group
        """
        # Construct the group and add metadata
        grp = Node.to_h5(self,group)

        # Add any attributes which are themselves emd nodes
        for k,v in vars(self).items():
            if isinstance(v,Node):
                if not isinstance(v,Root):
                    v.name = k
                    attr_grp = v.to_h5(grp)
                    attr_grp.attrs['emd_group_type'] = 'custom_' + \
                        attr_grp.attrs['emd_group_type']

        return grp






# This file defines a subclass of Array
# which contains a data array and a few attributes


from emdfile import Array, Metadata, _read_metadata
import numpy as np
from os.path import basename

class ArraySubclass(Array):
    """
    """

    def __init__(
        self,
        data,
        name = 'my_array_subclass',
        color = 'blue',
        number = 7,
        elephants = True
        ):
        """
        For some tuple `shape`, makes a data array of this
        shape and wraps it as an Array instance.
        """
        # Initialize as the parent class
        Array.__init__(
            self,
            name = name,
            data = data
        )

        # Assign attributes
        self.color = color
        self.number = number
        self.elephants = elephants

        pass



    # Read methods
    @classmethod
    def _get_constructor_args(cls,group):
        """
        Returns a dictionary of arguments to pass to __init__
        """
        # get array constructor args
        array_constr_args = Array._get_constructor_args(group)

        # get metadata
        md = _read_metadata(group,'_attr_metadata')

        # make the dictionary of constructor arguments and return
        constructor_args = {
            'name' : basename(group.name),
            'data' : array_constr_args['data'],
            'color' : md['color'],
            'number' : md['number'],
            'elephants' : md['elephants']
        }
        return constructor_args


    # Write methods
    def to_h5(self,group):
        """
        Store the extra class attributes as metadata before saving
        """
        self.metadata = Metadata(
            name = '_attr_metadata',
            data = {
                'color' : self.color,
                'number' : self.number,
                'elephants' : self.elephants,
            }
        )
        Array.to_h5(self,group)






import inspect
import types
import sys


# Define the EMD group types

EMD_base_group_types = (
    "root",
    "metadatabundle",
    "metadata",
)
EMD_data_group_types = (
    "node",
    "array",
    "pointlist",
    "pointlistarray",
    "custom",
)
EMD_custom_group_types = tuple(
    ["custom_"+s for s in EMD_data_group_types]
)
EMD_group_types = EMD_base_group_types + EMD_data_group_types + EMD_custom_group_types




def _get_class(grp):
    """
    Returns a dictionary of Class constructors from corresponding strings
    """
    from emdfile import classes

    # Build lookup table for classes
    lookup = {}
    for name, obj in inspect.getmembers(classes):
        if inspect.isclass(obj):
            lookup[name] = obj

    # hook for dependent package classes
    for module in _get_dependent_packages():
        _walk_module_find_classes(module, lookup)

    # Get the class from the group tags and return
    try:
        classname = grp.attrs['python_class']
        __class__ = lookup[classname]
        return __class__
    except KeyError:
        raise Exception(f"Unknown classname {classname}")




def _get_dependent_packages():
    """
    Searches packages with the top level attribute "_emd_hook" = True.
    Returns a generator of all such packages
    """
    for module in sys.modules.values():
        if isinstance(module, types.ModuleType):
            if hasattr(module, "_emd_hook"):
                if module._emd_hook is True:
                    yield module


def _walk_module_find_classes(mod, dic, depth=0, maxdepth=6):
    """
    Searches the tree of a Python module for emd classes, and
    populates dictionary dic with them
    """
    from emdfile import Node,Metadata
    if depth >= maxdepth:
        return
    for name, obj in inspect.getmembers(mod):
        if inspect.isclass(obj):
            if Node in obj.mro() or Metadata in obj.mro():
                dic[name] = obj
            else:
                pass
        elif inspect.ismodule(obj):
            if hasattr(obj, "_emd_hook"):
                if obj._emd_hook == True:
                    _walk_module_find_classes(
                        obj, dic, depth=depth+1, maxdepth=maxdepth)
        else:
            pass





from emdfile import PointList, Metadata, _read_metadata
import numpy as np

class PointListSubclass(PointList):
    """
    """

    def __init__(
        self,
        fields,
        length,
        name = 'my_points_subclass',
        ):
        """
        fields is a list of strings and length is an int.
        """

        # make the data
        data = np.zeros(
            length,
            dtype = [(x,np.float32) for x in fields]
        )

        # initialize
        PointList.__init__(
            self,
            name = name,
            data = data
        )

        pass



    # Read methods
    @classmethod
    def _get_constructor_args(cls,group):
        """
        Returns a dictionary of arguments to pass to __init__
        """
        # get array constructor args
        kwargs = PointList._get_constructor_args(group)
        name = kwargs['name']
        data = kwargs['data']

        # make the dictionary of constructor arguments and return
        constructor_args = {
            'fields' : data.dtype.names,
            'length' : len(data),
            'name' : name
        }
        return constructor_args


    def _populate_instance(self,group):
        """
        If we don't define this method, the class will write and read successfully -
        however, because the class __init__ method creates a new data array populated
        by zeros, a class instance loaded from an H5 *won't* point to the data array
        from H5.  (If this was all this class did, we probably wouldn't want the
        constructor to look like this for this reason - let's assume there's some
        reason it needs to be this way!). Adding this method re-assigns the data
        array to the data stored in the H5.
        """
        # get the data
        kwargs = PointList._get_constructor_args(group)
        data = kwargs['data']

        # re-assign .data
        self.data = data

        pass







# Demos use of the emd.Custom class


from typing import Optional,Union
import numpy as np
from os.path import basename

from emdfile import (
    Custom,
    Array,
    PointList,
    Metadata
)



class MyArrayClass(Array):
    """ A very simple Array subclass
    """

    def __init__(
        self,
        data,
        name = 'my_array_class'
        ):
        """ `data` is a numpy array
        """
        Array.__init__(self,name=name,data=data)

    # read
    @classmethod
    def _get_constructor_args(cls,group):
        """
        Returns a dictionary of args/values to pass to the class constructor
        """
        ar_constr_args = Array._get_constructor_args(group)
        args = {
            'data' : ar_constr_args['data'],
            'name' : ar_constr_args['name'],
        }
        return args


class MyPointsClass(PointList):
    """ A PointList subclass with specified dimension fields and metadata
    """

    def __init__(
        self,
        x,
        y,
        name = 'my_points_class'
        ):
        """ `x` and `y` are 1D arrays of the same length
        """
        data = np.empty(
            len(x),
            dtype = [
                ('x','>f8'),
                ('y','>f8'),
            ]
        )
        data['x'],data['y'] = x,y

        PointList.__init__(self,name=name,data=data)

        self.metadata = Metadata(
            name = 'metadata',
            data = {
                'length' : len(data)
            }
        )

    # read
    @classmethod
    def _get_constructor_args(cls,group):
        """
        Returns a dictionary of args/values to pass to the class constructor
        """
        pl_constr_args = PointList._get_constructor_args(group)
        args = {
            'x' : pl_constr_args['data']['x'],
            'y' : pl_constr_args['data']['y'],
            'name' : pl_constr_args['name'],
        }
        return args



class MyCustomClass(Custom):
    """ A class holding two Arrays and some metadata
    """

    def __init__(
        self,
        data,
        name = 'my_custom_class'
        ):
        """ `data` is a numpy array
        """
        Custom.__init__(self,name=name)


        # define some data arrays
        self.data = Array(
            data = data,
            name = 'data'
        )
        self.moredata = Array(
            data = np.tile(data,(2,2)),
            name = 'moredata'
        )


        # add some metadata
        self.metadata = Metadata(
            name = 'my_metadata',
            data = {
                'cats' : 'aregreat'
            }
        )



    # read
    @classmethod
    def _get_constructor_args(cls,group):
        """
        """
        # get EMD group data
        emd_data = cls._get_emd_attr_data(cls,group)

        kwargs = {
            'name' : basename(group.name),
            'data' : emd_data['data'].data
        }

        return kwargs






_emd_hook = True   # this hook is necessary!

from .array_subclass_sample1 import SimpleArraySubclass
from .array_subclass_sample2 import ArraySubclass
from .points_subclass_sample import PointListSubclass
from .raggedarray_subclass_sample import PointListArraySubclass
from .custom_subclass_sample import CustomSubclass






# Defines a class, PointList, for storing / accessing / manipulating data
# in the form of lists of vectors in named dimensions.  Wraps numpy
# structured arrays.

import numpy as np
import h5py
from typing import Optional
from os.path import basename

from emdfile.classes.tree import Node


class PointList(Node):
    """
    A wrapper around structured numpy arrays, with read/write functionality in/out of
    EMD formatted HDF5 files.

    """
    _emd_group_type = 'pointlist'

    def __init__(
        self,
        data: np.ndarray,
        name: Optional[str] = 'pointlist',
        ):
        """
		Instantiate a PointList.

        Args:
            data (structured numpy ndarray): the data; the dtype of this array will
                specify the fields of the PointList.
            name (str): name for the PointList

        Returns:
            a PointList instance
        """
        super().__init__()

        # populate data and metadata
        self.data = data
        self.name = name
        self._dtype = self.data.dtype
        self._fields = self.data.dtype.names
        if self._fields is not None:
            self._types = tuple([self.data.dtype.fields[f][0] for f in self.fields])
        else:
            self._fields = ('',)
            self._types = (self._dtype,)


    # properties
    @property
    def dtype(self):
        return self._dtype
    @dtype.setter
    def dtype(self, dtype):
        self._dtype = dtype

    @property
    def fields(self):
        return self._fields
    @fields.setter
    def fields(self, x):
        self.data.dtype.names = x
        self._fields = x

    @property
    def types(self):
        return self._types

    def __len__(self):
        return np.atleast_1d(self.data).shape[0]
    @property
    def length(self):
        return len(self)


    ## Add, remove, sort data

    def __add__(self, data):
        """
        Append a numpy structured array and return the concatenated pointlist.
        The dtypes must agree.
        """
        assert self.dtype == data.dtype, "Error: dtypes must agree"
        if isinstance(data,PointList):
            data = data.data
        ans = np.append(self.data, data)
        return PointList(
            name = self.name,
            data = ans
        )

    def add(self, data):
        """
        Appends a numpy structured array. Its dtypes must agree with the existing data.
        """
        self.data = (self + data).data

    def remove(self, mask):
        """ Removes points wherever mask==True
        """
        assert np.atleast_1d(mask).shape[0] == self.length, "deletemask must be same length as the data"
        inds = mask.nonzero()[0]
        self.data = np.delete(self.data, inds)

    def sort(self, field, order='ascending'):
        """
        Sorts the point list according to field,
        which must be a field in self.dtype.
        order should be 'descending' or 'ascending'.
        """
        assert field in self.fields
        assert (order=='descending') or (order=='ascending')
        if order=='ascending':
            self.data = np.sort(self.data, order=field)
        else:
            self.data = np.sort(self.data, order=field)[::-1]


    ## Copy, copy+modify PointList

    def copy(self, name=None):
        """ Returns a copy of the PointList. If name=None, sets to `{name}_copy`
        """
        name = name if name is not None else self.name+"_copy"

        pl = PointList(
            data = np.copy(self.data),
            name = name)

        for k,v in self.metadata.items():
            pl.metadata = v.copy(name=k)

        return pl

    def add_fields(self, new_fields, name=''):
        """
        Creates a copy of the PointList, but with additional fields given by new_fields.

        Args:
            new_fields: a list of 2-tuples, ('name', dtype)
            name: a name for the new pointlist
        """
        dtype = []
        for f,t in zip(self.fields,self.types):
            dtype.append((f,t))
        for f,t in new_fields:
            dtype.append((f,t))

        data = np.zeros(self.length, dtype=dtype)
        for f in self.fields:
            data[f] = np.copy(self.data[f])

        return PointList(data=data, name=name)

    def add_data_by_field(self, data, fields=None):
        """
        Add a list of data arrays to the PointList, in the fields
        given by `fields`. If `fields` is not specified, assumes the data
        arrays are in the same order as self.fields

        Args:
            data (list): arrays of data to add to each field
        """

        if data[0].ndim == 0:
            L = 1,
        else:
            L = data[0].shape[0]
        newdata = np.zeros(L,dtype=self.dtype)

        _fields = self.fields if fields is None else fields

        for d,f in zip(data, _fields):
            newdata[f] = d

        self.data = np.append(self.data,newdata)




    ## Representation to standard output
    def __repr__(self):

        space = ' '*len(self.__class__.__name__)+'  '
        string = f"{self.__class__.__name__}( A length {self.length} PointList called '{self.name}',"
        string += "\n"+space+f"with {len(self.fields)} fields:"
        string += "\n"
        space2 = max([len(field) for field in self.fields])+3
        for f,t in zip(self.fields,self.types):
            string += "\n"+space+f"{f}{(space2-len(f))*' '}({str(t)})"
        string += "\n)"

        return string


    # Slicing
    def __getitem__(self, v):
        return self.data[v]



    # HDF5 i/o

    # write
    def to_h5(self,group):
        """
        Takes an h5py Group instance and creates a subgroup containing
        this PointList, tags indicating its EMD type and Python class,
        and the pointlist's data and metadata.

        Accepts:
            group (h5py Group)

        Returns:
            (h5py Group) the new pointlist's group
        """
        # Construct group and add metadata
        grp = Node.to_h5(self,group)

        # Add data
        for f,t in zip(self.fields,self.types):
            group_current_field = grp.create_dataset(
                f,
                data = self.data[f]
            )
            group_current_field.attrs.create("dtype", np.string_(t))

        # Return
        return grp




    # read
    @classmethod
    def _get_constructor_args(cls,group):
        """
        Returns a dictionary of args/values to pass to the class constructor
        """
        # Get PointList metadata
        fields = list(group.keys())
        fields = [f for f in fields if isinstance(group[f],h5py.Dataset)]
        dtype = []
        for field in fields:
            curr_dtype = group[field].attrs["dtype"].decode('utf-8')
            dtype.append((field,curr_dtype))
        length = len(group[fields[0]])

        # Get data
        data = np.zeros(length,dtype=dtype)
        if length > 0:
            for field in fields:
                data[field] = np.array(group[field])

        # make args dictionary and return
        return {
            'data' : data,
            'name' : basename(group.name)
        }






from os import remove
from os.path import exists,join

from py4DSTEM import _TESTPATH as dirpath


# Set paths
single_node_path = join(dirpath,"test_h5_single_node.h5")
unrooted_node_path = join(dirpath,"test_h5_unrooted_node.h5")
whole_tree_path = join(dirpath,"test_h5_whole_tree.h5")
subtree_path = join(dirpath,"test_h5_subtree.h5")
subtree_noroot_path = join(dirpath,"test_h5_subtree_noroot.h5")
subtree_append_to_path = join(dirpath,"test_h5_append_to.h5")
subtree_append_over_path = join(dirpath,"test_h5_append_over.h5")


paths = [
    single_node_path,
    unrooted_node_path,
    whole_tree_path,
    subtree_path,
    subtree_noroot_path,
    subtree_append_to_path,
    subtree_append_over_path
]
for p in paths:
    if exists(p):
        remove(p)





# Class imports

from emdfile.classes.metadata import Metadata
from emdfile.classes.tree import Node,Root
from emdfile.classes.array import Array
from emdfile.classes.pointlist import PointList
from emdfile.classes.pointlistarray import PointListArray
from emdfile.classes.custom import Custom





import emdfile as emd
import numpy as np

# filepath
from os import getcwd, remove
from os.path import join, exists
path = join(getcwd(),"test.h5")



class TestListWrite():

    def setup_method(self):
        # delete the file
        if exists(path):
            remove(path)

    def teardown_method(self):
        # delete the file
        if exists(path):
            remove(path)


    # tests

    def test_two_unrooted_nodes(self):

        # make two arrays
        ar1 = emd.Array(
            data = np.array([[1,2],[3,4]]),
            name = 'an_array'
        )
        ar2 = emd.Array(
            data = np.arange(10),
            name = 'another_array'
        )

        # save them
        emd.save(
            filepath = path,
            data = [ar1,ar2],
        )

        # read them
        data = emd.read( path )

        # check
        assert(np.array_equal(data.tree('an_array').data, ar1.data))
        assert(np.array_equal(data.tree('another_array').data, ar2.data))


    def test_two_rooted_nodes(self):

        # make two arrays
        ar1 = emd.Array(
            data = np.eye(4),
            name = 'array1'
        )
        ar2 = emd.Array(
            data = np.arange(10),
            name = 'array2'
        )

        # add roots
        root1 = emd.Root(name='root1')
        root2 = emd.Root(name='root2')
        root1.tree(ar1)
        root2.tree(ar2)

        # save them
        emd.save(
            filepath = path,
            data = [ar1,ar2],
        )

        # read them
        data1 = emd.read(
            path,
            emdpath = 'root1'
        )
        data2 = emd.read(
            path,
            emdpath = 'root2'
        )

        # check
        assert(np.array_equal(data1.data, ar1.data))
        assert(np.array_equal(data2.data, ar2.data))


    def test_two_trees(self):

        # make several arrays
        ar1 = emd.Array(
            data = np.array([[1,2],[3,4]]),
            name = 'array1'
        )
        ar2 = emd.Array(
            data = np.arange(10),
            name = 'array2'
        )
        ar3 = emd.Array(
            data = np.arange(6),
            name = 'array3'
        )

        # add roots
        root1 = emd.Root(name='root1')
        root2 = emd.Root(name='root2')
        root1.tree(ar1)
        root2.tree(ar2)
        root2.tree(ar3)

        # save them
        emd.save(
            filepath = path,
            data = [root1,root2],
        )

        # read them
        data1 = emd.read(
            path,
            emdpath = 'root1'
        )
        data2 = emd.read(
            path,
            emdpath = 'root2'
        )


        # check
        assert(np.array_equal(data1.data, ar1.data))
        assert(np.array_equal(data2.tree('array2').data, ar2.data))
        assert(np.array_equal(data2.tree('array3').data, ar3.data))



    def test_mixed_list(self):
        """
        This test makes

        root1
          |--array1
          |--array2

        root2
          |--array3
          |--array4
          |--array5

        array6
        array7
        nparray
        dict

        and then calls

        emd.save(
            path,
            [
                root1,
                array3,
                array4,
                array6,
                array7,
                nparray,
                dict
            ]
        )

        which should save

        file
          |--root1
          |    |--array1
          |    |--array2
          |
          |--root2
          |    |--array3
          |    |--array4
          |
          |--root_savedlist
               |--array6
               |--array7
               |--nparray
               |--dict (metadata)

        Note that root2 should *not* contain array 5 in the saved file.
        """

        # make several arrays
        ar1 = emd.Array(
            data = np.array([[1,2],[3,4]]),
            name = 'array1'
        )
        ar2 = emd.Array(
            data = np.arange(10),
            name = 'array2'
        )
        ar3 = emd.Array(
            data = np.arange(6),
            name = 'array3'
        )
        ar4 = emd.Array(
            data = np.arange(24).reshape((2,3,4)),
            name = 'array4'
        )
        ar5 = emd.Array(
            data = np.arange(30).reshape((2,3,5)),
            name = 'array5'
        )
        ar6 = emd.Array(
            data = np.ones((2,4)),
            name = 'array6'
        )
        ar7 = emd.Array(
            data = np.eye(5),
            name = 'array7'
        )
        nparray1 = np.ones((2,3,2,3))
        nparray2 = np.ones((2,3,2,3,2,3))
        dic1 = {
            'x' : 1,
            'y' : True
        }
        dic2 = {
            'z' : -2,
        }

        # add roots
        root1 = emd.Root(name='root1')
        root1.tree(ar1)
        root1.tree(ar2)
        root2 = emd.Root(name='root2')
        root2.tree(ar3)
        root2.tree(ar4)
        root2.tree(ar5)

        # save them
        emd.save(
            filepath = path,
            data = [root1,ar3,ar4,ar6,ar7,nparray1,nparray2,dic1,dic2],
        )

        # TODO - remove
        #emd.save(
        #    filepath = path,
        #    data = root1
        #)
        #emd.save(
        #    filepath = path,
        #    data = root2,
        #    mode = 'a'
        #)
        #root_tmp = emd.Root(name='root_savedlist')
        #root_tmp.tree(ar6)
        #root_tmp.tree(ar7)
        #root_tmp.tree(
        #    emd.Array(
        #        name = 'nparray',
        #        data=nparray
        #    )
        #)
        #root_tmp.metadata = emd.Metadata(
        #    name = 'dictionary',
        #    data = dic
        #)
        #emd.save(
        #    filepath = path,
        #    data = root_tmp,
        #    mode = 'a'
        #)

        # read them
        data1 = emd.read(
            path,
            emdpath = 'root1'
        )
        data2 = emd.read(
            path,
            emdpath = 'root2'
        )
        data3 = emd.read(
            path,
            emdpath = 'root_savedlist'
        )


        # check
        assert(np.array_equal(
            data1.tree('array1').data,
            ar1.data
        ))
        assert(np.array_equal(
            data1.tree('array2').data,
            ar2.data
        ))
        assert(np.array_equal(
            data2.tree('array3').data,
            ar3.data
        ))
        assert(np.array_equal(
            data2.tree('array4').data,
            ar4.data
        ))
        assert(
            'array5' not in data2.treekeys
        )
        assert(np.array_equal(
            data3.tree('array6').data,
            ar6.data
        ))
        assert(np.array_equal(
            data3.tree('array7').data,
            ar7.data
        ))
        assert(np.array_equal(
            data3.tree('array_0').data,
            nparray1
        ))
        assert(np.array_equal(
            data3.tree('array_1').data,
            nparray2
        ))
        assert(
            'dictionary_0' in data3.metadata.keys()
        )
        assert(
            'dictionary_1' in data3.metadata.keys()
        )
        d = data3.metadata['dictionary_0']
        for k in d.keys:
            assert(k in dic1.keys())
            assert(d[k] == dic1[k])
        d = data3.metadata['dictionary_1']
        for k in d.keys:
            assert(k in dic2.keys())
            assert(d[k] == dic2[k])






from emdfile import PointList, Metadata, _read_metadata
import numpy as np
from os.path import basename

class PointListSubclass_TEMPLATE(PointList):
    """
    An emdfile pointlist subclass template.

    There are two required methods: __init__, and _get_constructor_args.

    The __init__ method should call PointList.__init__.

    _get_constructor_args is required for reading the class.  When the
    emdfile read() function finds an H5 group representing this class,
    this method is called.  It should collect and return a dictionary
    representing arguments that will be passed to the __init__ method.

    There are two optional methods: _populate_instance and to_h5.

    If _populate method is defined then run when reading this object from
    an H5 file, this method is run after instantiation, enabling additional
    setup or configuration.

    emdfile classes already have a .to_h5 method defined, which will handle
    storing the data, the emdfile class' native metadata (e.g its name, dim
    vectors, fields, shapes), and metadata stored in the .metadata property. If
    additional customization is desired, the .to_h5 method can be overwritten.
    In this case, the parent class' .to_h5 method should be run to save the
    normal data and metadata.
    """

    def __init__(
        self,
        name = 'my_pointlist_subclass'
        *args,
        **kwargs
        ):
        """
        Required. Should call PointList.__init__.
        """

        # code
        # ...
        # ...

        # Initialize as the parent class
        PointList.__init__(
            self,
            name = name,
            data = data
        )

        # code
        # ...
        # ...

        pass



    # Read methods


    # This method is required.

    @classmethod
    def _get_constructor_args(cls,group):
        """
        Required. Must return a dictionary representing arguments which will
        be passed to __init__ at read time.
        """

        # Retrieve PointList constructor args
        pointlist_constr_args = PointList._get_constructor_args(group)

        # Retrieve Metadata dictionaries
        some_metadata = _read_metadata(group,'name')

        # code
        # ...
        # ...

        # make the dictionary of constructor arguments
        constructor_args = {
            'name' : basename(group.name),
            'data' : pointlist_constr_args['data'],
            'thing' : some_metadata['thing']
            # etc.
        }

        # return
        return constructor_args



    # This method is optional

    def _populate_instance(self,group):
        """
        Optional.  During read, this method is run after object instantiation.
        """
        pass




    # Write methods

    # This method is optional

    def to_h5(self,group):
        """
        Optional. If defined, should call PointList.to_h5(self,group) to handle
        saving the parent class' data/metadata.
        """
        # Here some class attributes are stored in a Metadata dict which is
        # placed in the .metadata propetry, such that it's saved in the 
        # subsequent `to_h5` call
        self.metadata = Metadata(
            name = '_attr_metadata',
            data = {
                'x' : self.x,
                'y' : self.y
            }
        )
        PointList.to_h5(self,group)






# Write EMD 1.0 formatted HDF5 files.

import h5py
import numpy as np
from warnings import warn
from os.path import exists,basename
from os import remove
from uuid import uuid4
from emdfile.read import _is_EMD_file, _get_EMD_rootgroups
from emdfile.classes.utils import EMD_data_group_types
from emdfile.classes import (
    Node,
    Root,
    Array,
    Metadata
)


def write(
    filepath,
    data,
    mode = 'w',
    emdpath = None,
    tree = True,
    ):
    """
    Saves data to a .h5 file at filepath.

    Calling

        >>> save(path, data)

    if `data` is a Root instance saves this root and its entire tree to a new
    file. If `data` is any other type of rooted node (i.e. a node inside of
    some runtime data tree), this code writes a new file with a single tree
    using this node's root (even if this node is far downstream of the root
    node), placing this node and the tree branch underneath it inside that
    root. In both cases, the root metadata is stored in the new H5 root node.
    If `data` is an unrooted node (i.e. a freestanding node not connected to
    a tree), this code creates a new root node with no metadata and this node's
    name, and places this node inside that root in a new file.

    If `data` is a numpy array or Python dictionary, wraps data in either an
    emd.Array or emd.Metadata instance, assigns the name 'np.array' or
    'dictionary', places the object in a root of this name and saves. If
    `data` is a list of objects which are all numpy arrays, Python dictionaries,
    or emd.Node instances, places all these objects into a single root, assigns
    the roots name according to the first object in the list, and saves.

    To write a single node from a tree,  set `tree` to False. To write the
    tree underneath a node but exclude the node itself set `tree` to None.

    To add to an existing EMD file, use the `mode` argument to set append or
    appendover mode. If the `emdpath` variable is not set and `data` has a
    runtime root that does not exist in the EMD root groups already present,
    adds the new root and writes as described above. If `emdpath` is not set
    and the runtime root group matches a root group that's already present,
    this function performs a diff operation between the root metadata and
    data nodes from `data` and those already in the H5 file. Append mode adds
    any data/metadata groups with no equivalent (i.e. same name and tree
    location) in the H5 tree, while skipping any data/metadata already found
    in the tree. Appendover adds any data/metadata with no equivalent already
    in the H5 tree, and overwrites any data/metadata groups that are already
    represented in the HDF5 with the new data. Note that this function does
    not attempt to take a diff between the contents of the groups and the
    runtime data groups - it only considers the names and their locations in
    the tree. If append or appendover mode are used and filepath is set to
    a location that does not already contain a file on the filesystem,
    behavior is identical to write mode. When appendover mode overwrites
    data, it is erasing the old links and creating new links to new data;
    however, the HDF5 file does not release the space on the filesystem.
    To free up storage, set mode to 'appendover', and this function will
    add a final step to re-write then delete the old file.

    The `emdpath` argument is used to append to a specific location in an
    extant EMD file downstream of some extant root. If passed, it must point
    to a valid location in the EMD file. This function will then perform a
    diff and write as described in the prior paragraph, except beginning
    from the H5 node specified in `emdpath`. Note that in this case the root
    metadata is still compared to and added or overwritten in the H5 root node,
    even if the remaining data is being added to some downstream branch.


    Args:
        filepath: path where the file will be saved
        data: an EMD data class instance
        mode (str): supported modes and their keys are:
                - write ('w','write')
                - overwrite ('o','overwrite')
                - append ('a','+','append')
                - appendover ('ao','oa','o+','+o','appendover')
            Write mode writes a new file, and raises an exception if a file
            of this name already exists.  Overwrite mode deletes any file of
            this name that already exists and writes a new file. Append and
            appendover mode write a new file if no file of this name exists,
            or if a file of this name does exist, adds new data to the file.
            The specific behavior of append and appendover depend on the
            `data`,`emdpath`, and `tree` arguments as discussed in more detail
            above. Broadly, both modes attempt to detemine the difference
            between the data passed and that present in the extent HDF5 file
            tree, add any data not already in the H5, and then either skips
            or overwrites conflicting nodes in append or appendover mode,
            respectively.
        tree: indicates how the object tree nested inside `data` should
            be treated.  If `True` (default), the entire tree is saved.
            If `False`, only this object is saved, without its tree. If
            `None`, saves the entire tree underneath `data`, but not
            the node at `data` itself.
        emdpath (str or None): optional parameter used in conjunction with
            append or appendover mode; if passed in write or overwrite mode,
            this argument is ignored. Indicates where in an existing EMD
            file tree to place the data. Must be a '/' delimited string
            pointing to an existing EMD file tree node.
    """
    # parse mode
    writemode = [
        'w',
        'write'
    ]
    overwritemode = [
        'o',
        'overwrite'
    ]
    appendmode = [
        'a',
        '+',
        'append'
    ]
    appendovermode = [
        'oa',
        'ao',
        'o+',
        '+o',
        'appendover'
    ]
    allmodes = writemode + overwritemode + appendmode + appendovermode

    # emdpath implies append mode
    if emdpath is not None and mode not in appendovermode:
        mode = 'a'


    # validate `mode` and `tree` inputs
    er = f"unrecognized mode {mode}; mode must be in {allmodes}"
    assert(mode in allmodes), er
    if tree == 'noroot':
        warn("`tree = 'noroot'` is deprecated and will be removed in a future version. Use `tree = None` instead.")
        tree = None
    assert(tree in (True,False,None)), f"invalid value {tree} passed for `tree`"
    if mode in writemode:
        assert(not(exists(filepath))), "A file already exists at this destination; use append or overwrite mode, or choose a new file path."


    # validate `data` inputs, and
    # handle non-Node `data` inputs

    # numpy array -> Array
    if isinstance(data, np.ndarray):
        root = Root(name='root')
        data = Array(name='np.array',data=data)
        root.add_to_tree(data)

    # dictionaries -> Metadata
    elif isinstance(data, dict):
        root = Root(name='root')
        md = Metadata(name='dictionary',data=data)
        root.metadata = md
        data = root

    # Metadata
    elif isinstance(data, Metadata):
        root = Root(name='root')
        root.metadata = data
        data = root



    ### Saving lists of items ###

    elif isinstance(data, (list,tuple)):
        assert(all( [isinstance(x,(np.ndarray,dict,Node)) for x in data] )), \
            "can only save np.array, dictionary, or emd.Node objects"

        # prepare roots

        # sort items into roots, nodes, and (np.array/dict)s
        list_roots = []
        list_nodes = []
        list_others = []
        for x in range(len(data)):
            x = data.pop(0)
            if isinstance(x,Root):
                list_roots.append(x)
            elif isinstance(x,Node):
                list_nodes.append(x)
            else:
                list_others.append(x)

        # sort nodes+array/dicts into rooted and unrooted
        list_rooted_nodes = []
        list_unrooted_items = []
        for x in list_nodes:
            if x.root is None:
                list_unrooted_items.append(x)
            else:
                list_rooted_nodes.append(x)
        list_unrooted_items += list_others

        # place all unrooted items into a single root
        if len(list_unrooted_items) > 0:
            root_savedlist = Root(name = "root_savedlist")
            ind_ars = 0
            ind_dics = 0
            for x in list_unrooted_items:
                if isinstance(x,Node):
                    root_savedlist.tree(x)
                elif isinstance(x,np.ndarray):
                    ar = Array(
                        name = f"array_{ind_ars}",
                        data = x
                    )
                    root_savedlist.tree(ar)
                    ind_ars += 1
                else:
                    dic = Metadata(
                        name = f"dictionary_{ind_dics}",
                        data = x
                    )
                    root_savedlist.metadata = dic
                    ind_dics += 1
            list_roots.insert(0,root_savedlist)


        # sort through rooted nodes, grouping them by roots
        # ensure only selected items from each root are saved
        # trees holding several selected nodes will be flattened
        dict_roots = {}
        if len(list_rooted_nodes) > 0:
            # list all roots, check for duplicate names
            for x in list_rooted_nodes:
                rcurr,rname = x.root,x.root.name
                if rname not in dict_roots.keys():
                    dict_roots[rname] = rcurr
                else:
                    assert(rcurr is dict_roots[rname]), f"Two nodes have different roots with identical names! Try changing one of their names."
            # make new roots, replace in the dict
            for rname,rcurr in dict_roots.items():
                root_new = Root( name=rname )
                for m in rcurr.metadata.values():
                    root_new.metadata = m
                dict_roots[rname] = root_new

        # save

        # set mode to append
        if mode in writemode:
            mode = 'a'
        elif mode in overwritemode:
            if exists(filepath):
                remove(filepath)
            mode = 'a'
        else:
            pass

        # write roots
        for root in list_roots:
            write(
                filepath,
                root,
                tree = True,
                mode = mode
            )

        # write nodes
        for root in dict_roots.values():
            write(
                filepath,
                root,
                mode = mode
            )
        for item in list_rooted_nodes:
            write(
                filepath,
                item,
                emdpath = item.root.name,
                tree = False,
                mode = 'ao'
            )
        return



    ### Saving emd Node data ###

    # `data` should now be a Node!
    assert(isinstance(data,Node)), f"invalid type {type(data)} found for `data`"

    # handle rootless data
    added_a_root = False
    if data._root is None:
        added_a_root = True
        root = Root(name=data.name+"_root")
        root.add_to_tree(data)
    else:
        # get the root
        root = data.root



    # overwrite mode - delete existing file
    if mode in overwritemode:
        if exists(filepath):
            remove(filepath)
        mode = 'w'



    # write a new file
    if mode in writemode or (
        mode in appendmode+appendovermode and not exists(filepath)):


        # open the file
        with h5py.File(filepath, 'w') as f:

            # write header
            _write_header(
                file = f
            )

            # write the file
            _write_from_root(
                file = f,
                root = root,
                data = data,
                tree = tree
            )



    # append to an existing file
    else:

        # validate that its an EMD file
        # get the rootgroups
        assert(_is_EMD_file(filepath)), "{filepath} does not point to an EMD 1.0 file"
        emd_rootgroups = _get_EMD_rootgroups(filepath)

        # open the file
        with h5py.File(filepath, 'a') as f:



            # if the root doesn't already exist and emdpath is None,
            # do a simple write as above
            if not(root.name in emd_rootgroups) and (emdpath is None):

                _write_from_root(
                    file = f,
                    root = root,
                    data = data,
                    tree = tree
                )



            # if the root doesn't already exist and emdpath is specified,
            # append the data to the target node
            elif not(root.name in emd_rootgroups):

                # parse emdpath
                if emdpath[0] == '/':
                    emdpath = emdpath[1:]
                l = emdpath.split('/')
                rootname = l[0]
                treepath = '/'.join(l[1:])

                # get the rootgroup
                assert(rootname in f.keys()), f"No root called {rootname} found - check your `emdpath`"
                rootgroup = f[rootname]

                # validate the emdpath
                # set target_grp to targeted EMD node
                where = _validate_treepath(
                    rootgroup,
                    treepath
                )
                if where is False:
                    raise Exception(f"No node found at {emdpath} in the EMD tree called {rootname} - check your `emdpath`")
                elif where[1] is False:
                    raise Exception(f"No node found at {emdpath} in the EMD tree called {rootname} - check your `emdpath`")
                else:
                    target_grp = where[0]


                # append to the tree...

                # ...if data is Root and tree is False
                if isinstance(data,Root) and (tree is False):
                    raise Exception("Incompatible inputs: if appending from a Root to an existing tree, `tree` can't be False.  Try changing `data` or `tree`.")

                # ...if data is Root and tree is True or None
                elif isinstance(data,Root):
                    _write_tree(
                        target_grp,
                        data
                    )

                # ...if data is a Node and tree is False
                elif tree is False:
                    _write_single_node(
                        target_grp,
                        data
                    )

                # ...if data is a Node and tree is True
                elif tree is True:
                    target_grp = _write_single_node(
                        target_grp,
                        data
                    )
                    _write_tree(
                        target_grp,
                        data
                    )

                # ...if data is a Node and tree is None
                else:
                    _write_tree(
                        target_grp,
                        data
                    )



            # if the root does exist and emdpath is None,
            # peform diffmerge A
            elif emdpath is None:

                # choose how to handle conflicts
                appendover = True if mode in appendovermode else False

                # get the rootgroup
                rootgroup = f[root.name]

                # compare/append root metadata
                _append_root_metadata(
                    rootgroup = rootgroup,
                    root = root,
                    appendover = appendover
                )


                # choose behavior and write...
                if data is root:
                    # ...if the data is the root
                    if tree is True:
                        _append_branch(
                            rootgroup,
                            data,
                            appendover
                        )
                    else:
                        pass

                else:
                    where = _validate_treepath(
                        rootgroup,
                        data._treepath
                    )
                    # ...if the datapath is not in the H5 path
                    if where is False:
                        raise Exception("The data passed can't be added to it's corresponding H5 tree - check that the data's `_treepath` is present in the existing EMD file")
                    else:
                        where,inside = where
                        # ...if the datapath is in the H5 path
                        if inside is True:
                            if tree is True:
                                if appendover:
                                    next_node = _overwrite_single_node(
                                        where,
                                        data
                                    )
                                else:
                                    next_node = where
                                _append_branch(
                                    next_node,
                                    data,
                                    appendover
                                )
                            elif tree is False:
                                if appendover:
                                    next_node = _overwrite_single_node(
                                        where,
                                        data
                                    )
                                else:
                                    pass
                            else:
                                _append_branch(
                                    where,
                                    data,
                                    appendover
                                )
                        # ...if the datapath is one node beyond the H5 path
                        else:
                            if tree is True:
                                new_node = _write_single_node(
                                    where,
                                    data
                                )
                                _write_tree(
                                    new_node,
                                    data
                                )
                            elif tree is False:
                                _write_single_node(
                                    where,
                                    data
                                )
                                pass
                            else:
                                _write_tree(
                                    where,
                                    data
                                )



            # if the root does exist and emdpath is specified,
            # peform diffmerge B
            else:

                # choose how to handle conflicts
                appendover = True if mode in appendovermode else False

                # parse emdpath
                if emdpath[0] == '/':
                    emdpath = emdpath[1:]
                l = emdpath.split('/')
                rootname = l[0]
                treepath = '/'.join(l[1:])

                # get the rootgroup
                rootgroup = f[root.name]

                # validate the emdpath
                # set target_grp to targeted EMD node
                where = _validate_treepath(
                    rootgroup,
                    treepath
                )
                if where is False:
                    raise Exception(f"No node found at {emdpath} in the EMD tree called {rootname} - check your `emdpath`")
                elif where[1] is False:
                    raise Exception(f"No node found at {emdpath} in the EMD tree called {rootname} - check your `emdpath`")
                else:
                    target_grp = where[0]


                # compare/append root metadata
                _append_root_metadata(
                    rootgroup = rootgroup,
                    root = root,
                    appendover = appendover
                )



                # choose behavior and write...

                # ...if the data is the root
                if data is root:

                    # Confirm that the target node is downstream of the root...
                    assert(rootgroup.__contains__(target_grp.name)), "Specified target node not found in the EMD file - check your emdpath."

                    # get the path from source to target, then
                    # move `data` to the target node point
                    path_to_target = target_grp.name.replace(rootgroup.name,'')[1:]
                    try:
                        data = data.tree(path_to_target)
                    except AssertionError:
                        raise Exception("Append failure - the target EMD node exists downstream of the source EMD node, however the target is not present in the corresponding runtime tree")
                    # write
                    if appendover and tree in (True,False):
                        target_grp = _overwrite_single_node(
                            target_grp,
                            data
                        )
                    if tree in (True,None):
                        _append_branch(
                            target_grp,
                            data,
                            appendover
                        )


                # ...if the data is a node...
                else:
                    # validate the source node path
                    where = _validate_treepath(
                        rootgroup,
                        data._treepath
                    )
                    # ...if the source node is not in the H5
                    if where is False:
                        raise Exception("The data passed can't be appended to it's corresponding H5 tree - the source runtime node can't be matched to the existing tree")
                    else:
                        source_grp,inside = where

                        # ...if the source node is one node beyond the H5
                        if inside is False:
                            # ...if it is one node past the targetted node, write
                            if source_grp.name == target_grp.name:
                                if tree in (True,None):
                                    _append_branch(
                                        target_grp,
                                        data,
                                        appendover
                                    )
                                else:
                                    _write_single_node(
                                        target_grp,
                                        data
                                    )
                            # ...otherwise, raise an Exception
                            else:
                                raise Exception("The data passed can't be added to it's corresponding H5 tree - check that the data's `.tree()` path is present in the existing EMD file")

                        # ...if the source node is in inside the H5
                        else:
                            # ...if the source node is the target node, write
                            if source_grp.name == target_grp.name:
                                if appendover and tree in (True,False):
                                    target_grp = _overwrite_single_node(
                                        target_grp,
                                        data
                                    )
                                if tree in (True,None):
                                    _append_branch(
                                        target_grp,
                                        data,
                                        appendover
                                    )
                            # ...if the source node is one node downstream of the target, write
                            elif basename(source_grp.name) in list(target_grp.keys()):
                                target_grp = source_grp
                                if appendover and tree in (True,False):
                                    target_grp = _overwrite_single_node(
                                        target_grp,
                                        data
                                    )
                                if tree in (True,None):
                                    _append_branch(
                                        target_grp,
                                        data,
                                        appendover
                                    )
                            # ...if the target node is downstream of the source node...
                            elif source_grp.__contains__(target_grp.name):
                                # get the path from source to target, then
                                # move `data` to the target node point
                                path_to_target = target_grp.name.replace(source_grp.name,'')[1:]
                                try:
                                    data = data.tree(path_to_target)
                                except AssertionError:
                                    raise Exception("Append failure - the target EMD node exists downstream of the source EMD node, however the target is not present in the corresponding runtime tree")
                                # write
                                if appendover and tree in (True,False):
                                    target_grp = _overwrite_single_node(
                                        target_grp,
                                        data
                                    )
                                if tree in (True,None):
                                    _append_branch(
                                        target_grp,
                                        data,
                                        appendover
                                    )

                            # ...otherwise raise an exception
                            else:
                                raise Exception("Append failure - target node may not be downstream of source node.  Check the emdpath and the runtime data tree.")




    # if a root was added, remove it
    if added_a_root:
        data._root = None


    # end
    pass






# Utilities

def _write_header(
    file
    ):
    from emdfile import _PROGRAM_NAME, _USER_NAME
    file.attrs.create("emd_group_type",'file')
    file.attrs.create("version_major",1)
    file.attrs.create("version_minor",0)
    #file.attrs.create("version_release",0)
    file.attrs.create("UUID",str(uuid4()))
    file.attrs.create("authoring_program",_PROGRAM_NAME)
    file.attrs.create("authoring_user",_USER_NAME)


def _write_from_root(
    file,
    root,
    data,
    tree
    ):
    """ From an open h5py File with an EMD 1.0 header, adds a new root
    and data tree
    """
    # write the root
    rootgroup = _write_single_node(
        group = file,
        data = root,
    )
    rootgroup.attrs['emd_group_type'] = 'root'

    # write the rest
    if data is root:
        if tree is False:
            pass
        else:
            _write_tree(
                group=rootgroup,
                data=data
            )
    else:
        if tree is False:
            grp = _write_single_node(
                group = rootgroup,
                data = data
            )
        elif tree is True:
            grp = _write_single_node(
                group = rootgroup,
                data = data
            )
            _write_tree(
                group = grp,
                data = data
            )
        else:
            _write_tree(
                group = rootgroup,
                data = data
            )


def _write_single_node(
    group,
    data
    ):
    grp = data.to_h5(group)
    return grp


def _write_tree(
    group,
    data
    ):
    """ Writes the data tree underneath `data`; does not write `data`
    """
    for k in data._branch.keys():
        grp = _write_single_node(
            group = group,
            data = data._branch[k]
        )
        _write_tree(
            grp,
            data._branch[k]
        )


def _append_root_metadata(
    rootgroup,
    root,
    appendover
    ):
    # Determine if there is new group metadata
    if len(root._metadata)==0:
        return
    # Get file root metadata groups
    metadata_groups = []
    if "metadatabundle" not in rootgroup.keys():
        mdbundle_group = rootgroup.create_group('metadatabundle')
    else:
        mdbundle_group = rootgroup['metadatabundle']
        for k in mdbundle_group.keys():
            if "emd_group_type" in mdbundle_group[k].attrs.keys():
                if mdbundle_group[k].attrs["emd_group_type"] == "metadata":
                    metadata_groups.append(k)
    # loop
    for key in root._metadata:
        # if this group already exists
        if key in metadata_groups:
            # overwrite it
            if appendover:
                del(mdbundle_group[key])
                root._metadata[key].to_h5(mdbundle_group)
            # or skip it
            else:
                pass
        # otherwise, write it
        else:
            root._metadata[key].to_h5(mdbundle_group)
    return


def _validate_treepath(
    rootgroup,
    treepath
    ):
    """
    Accepts a file rootgroup and a runtime `treepath` string.
    If the treepath is not in the file, returns False.
    If the treepath is in the file, returns (grp, True) and
    if the treepath is one node beyond the file, returns (grp, False),
    where `grp` is the final h5py Group on treepath in the file tree.
    """
    grp_names = treepath.split('/')
    try:
        grp_names.remove('')
    except ValueError:
        pass
    group = rootgroup
    for i,name in enumerate(grp_names):
        if name not in group.keys():
            # catch for being one node beyond
            if i == len(grp_names)-1:
                return group, False
            return False
        group = group[name]
        try:
            assert(isinstance(group,h5py.Group))
        except AssertionError:
            return False
    return group, True


def _overwrite_single_node(
    group,
    data
    ):
    # get names
    groupname = group.name.split('/')
    name = groupname[-1]
    rootname = '/'+groupname[1]
    groupname = '/'+'/'.join(groupname[2:])

    # Validate
    assert(data.name == name), f"Can't overwrite - data/group names don't match: {data.name} != {name}"
    assert(groupname == data._treepath), f"Can't overwrite - data/group paths dont match: {group.name != data._treepath}"

    # Get parent group
    parentpath = data._treepath.split('/')
    parentpath = rootname+'/'.join(parentpath[:-1])
    parentgroup = group.file[parentpath]

    # Rename the old group
    parentgroup.move(name,"_tmp_"+name)

    # Write the new data 
    new_group = _write_single_node(
        parentgroup,
        data
    )

    # Copy the links
    keys = [k for k in group.keys() if "emd_group_type" in group[k].attrs.keys()]
    keys = [k for k in keys if group[k].attrs["emd_group_type"] in EMD_data_group_types]
    for key in keys:
        new_group[key] = group[key]

    # Remove the old group
    del(parentgroup["_tmp_"+name],group)

    # Return
    return new_group


def _append_branch(
    group,
    data,
    appendover
    ):
    groupkeys = [k for k in group.keys() if "emd_group_type" in group[k].attrs.keys()]
    # for each node under `data`...
    for key in data._branch.keys():
        d = data._branch[key]
        # ...if this node doesn't exist in the H5, do a simple write
        if d.name not in groupkeys:
            _write_single_node(
                group,
                d
            )
            _write_tree(
                group,
                d
            )
        # otherwise, overwrite or skip it, then call this fn again
        else:
            if appendover:
                next_node = _overwrite_single_node(
                    group[key],
                    d
                )
            else:
                next_node = group[key]
            _append_branch(
                next_node,
                d,
                appendover
            )





import emdfile as emd
import numpy as np


# import the classes inheriting from emdfile.
# `sample_custom_emd_classes` must be a python module
# which sets a variable `_emd_hook` to True in its
# toplevel namespace (in this example, in __init__.py)

from sample_custom_emd_classes import (
    SimpleArraySubclass,
    ArraySubclass,
    PointListSubclass,
    PointListArraySubclass,
    #MyCustomClass
)


# set a filepath
filepath = "/Users/Ben/Desktop/test.h5"


# run as a script...
if __name__ == "__main__":

    # instantiate classes
    array1 = SimpleArraySubclass(
        name = 'test_simple_array',
        data = np.zeros((2,2))
    )
    array2 = ArraySubclass(
        name = 'test_array',
        data = np.ones((2,3)),
        color = 'red',
        number = 3,
        elephants = False
    )
    pointlist = PointListSubclass(
        name = 'test_points',
        fields = ['x','y','z'],
        length = 8
    )
    pointlist['x'][1:] = 7
    pointlistarray = PointListArraySubclass(
        name = 'test_pointlistarray',
        shape = (4,5)
    )
    # populate with data
    for (x,y) in np.ndindex((4,5)):
        pointlistarray[x,y] += np.full(x,y,dtype=np.float32)
    # TODO custom

    # save
    emd.save(
        filepath,
        [
            array1,
            array2,
            pointlist,
            pointlistarray,
            #custom_instance
        ],
        mode = 'o'
    )

    # load
    data = emd.read(filepath)

    # display
    print("Loaded data:")
    print(data)
    print()
    print("Loaded tree:")
    data.tree()
    print()
    print("Simple Array subclass:")
    ar = data.tree('test_simple_array')
    print(ar)
    print(ar.name)
    print(ar.data)
    print()
    print("Array subclass with attributes:")
    ar = data.tree('test_array')
    print(ar)
    print(ar.name)
    print(ar.data)
    print(ar.color, ar.number, ar.elephants)
    print()
    print("PointList subclass:")
    pts = data.tree('test_points')
    print(pts)
    print(pts.data)
    print(pts.data.dtype)
    print()
    print("PointListArray subclass:")
    pla = data.tree('test_pointlistarray')
    print(pla)
    print(pla.dtype, pla.fields, pla.types)
    print(pointlistarray[2,3].data)
    print(pla[2,3].data)
    print()
    #print(data.tree('test_custom').data)







from typing import Optional
from os.path import basename
import h5py

from emdfile.classes import Metadata
from emdfile.classes.utils import EMD_group_types, _get_class


class Node:
    """
    Nodes contain attributes and methods paralleling
    the EMD 1.0 file specification in Python runtime objects.

    EMD 1.0 is a singly-rooted file format.  That is to say:
    An EMD data object can and must exist in one and only one
    EMD tree. An EMD file can contain any number of EMD trees, each
    containing data and metadata which is, within the limits of
    the EMD group specifications, of some arbitrary complexity.
    An EMD 1.0 file thus represents, stores, and enables
    access to some arbitrary data in long term storage on a file
    system in the form of an HDF5 file.  The Node class provides
    machinery for building trees of data and metadata which mirror
    the EMD tree format but which exist in a live Python instance,
    rather than on the file system. This facilitates ease of
    transfer between Python and the file system.

    Nodes are intended to be used a base class on which other, more
    complex classes can be biult. Nodes themselves contain the
    machinery for managing a tree heirarchy of other Nodes and
    Metadata instances, and for reading and writing those trees.
    They do not contain any particular data. Classes storing data
    and analysis methods which inherit from Node will inherit its
    tree management and EMD i/o functionality.

    Below, the 4 elements of the node class are each described in turn:
    roots, trees, metadata, and i/o.


    ROOTS

    EMD data objects can and must exist in one and only one EMD tree,
    each of which must have a single, named root node. To parallel this in
    our runtime objects, each Node has a root property, which can be found
    by calling `self.root`.

    By default new nodes have their root set to None. If a node
    with `.root == None` is saved to file, it is placed inside a
    new root with the same name as the object itself, and this
    is then saved to the file as a new (minimal) EMD tree.

    A new root node can be instantiated by calling

        >>> rootnode = Root(name=some_name).

    Objects added to an existing rooted tree (including a new root node)
    automatically have their root assigned to the root of that tree.
    Adding objects to trees is discussed below.


    TREES

    The tree associated with a node can be manipulated with the .tree
    method. If we have some rooted node `node1` and some unrooted node
    `node2`, the unrooted node can be added to the existing tree as a
    child of the rooted node with

        >>> node1.tree(node2)

    If we have a rooted node `node1` and another rooted node `node2`,
    we can't simply add node2 with the code above, as this would
    create a conflict between the two roots.  In this case, we can
    move node2 from its current tree to the new tree using

        >>> node1.tree(graft=node2)

    The .tree method has various additional functionalities, including
    printing the tree, retrieving objects from the tree, and cutting
    branches from the tree.  These are summarized below:

        >>> .tree()             # show tree from current node
        >>> .tree(show=True)    # show from root
        >>> .tree(show=False)   # show from current node
        >>> .tree(add=node)     # add a child node
        >>> .tree(get='path')   # return a '/' delimited child node
        >>> .tree(get='/path')  # as above, starting at root
        >>> .tree(cut=True)     # remove/return a branch, keep root metadata
        >>> .tree(cut=False)    # remove/return a branch, discard root md
        >>> .tree(cut='copy')   # remove/return a branch, copy root metadata
        >>> .tree(graft=node)   # remove/graft a branch, keep root metadata
        >>> .tree(graft=(node,True))    # as above
        >>> .tree(graft=(node,False))   # as above, discard root metadata
        >>> .tree(graft=(node,'copy'))  # as above, copy root metadata

    The show, add, and get methods can be accessed directly with

        >>> .tree(arg)

    for an arg of the appropriate type (bool, Node, and string), i.e.
    in most cases, the keyword can be dropped.  So

        >>> .tree()
        >>> .tree(node)
        >>> .tree(True)
        >>> .tree('some/node')

    will, respectively, print the tree from the current node to screen,
    add the node `node` to the tree, pring the tree from the root node
    to screen, and return the node at the emdpath 'some/node'.

    If a node needs to be added to a tree and it may or may not
    already have its own root, calling

        >>> .tree(add=node, force=True)

    or

        >>> .tree(node, force=True)

    will add the node to the tree, using a simple add if node has no
    root, and grafting it if it does have a root.


    METADATA

    Nodes can contain any number of Metadata instances, each of which
    wraps a Python dictionary of some arbitrary complexity (to within
    the limits of the Metadata group EMD specification, which limits
    permissible values somewhat).

    The code:

        >>> md1 = Metadata(name='md1')
        >>> md2 = Metadata(name='md2')
        >>> <<<  some code populating md1 + md2 >>>
        >>> node.metadata = md1
        >>> node.metadata = md2

    will create two Metadata objects, populate them with data, then
    add them to the node.  Note that Node.metadata is *not* a Python
    attribute, it is specially defined property, such that the last
    line of code does not overwrite the line before it - rather,
    assigning to the .metadata property adds the new metadata object
    to a running dictionary of arbitrarily many metadata objects.
    Both of these two metadata instances can therefore still be
    retrieved, using:

        >>> x = node.metadata['md1']
        >>> y = node.metadata['md2']

    Note, however, that if the second metadata instance has an identical
    name to the first instance, then in *will* overwrite the old instance.


    I/O

    # TODO

    """
    _emd_group_type = 'node'

    def __init__(
        self,
        name: Optional[str] = 'node'
        ):
        self.name = name
        self._branch = Branch()     # enables accessing child groups
        self._treepath = None   # enables accessing parent groups
        self._root = None
        self._metadata = {}


    @property
    def root(self):
        return self._root


    @property
    def metadata(self):
        return self._metadata
    @metadata.setter
    def metadata(self,x):
        assert(isinstance(x,Metadata))
        self._metadata[x.name] = x

    @property
    def treekeys(self):
        return self._branch.keys()


    # displays top level contents of the node
    def __repr__(self):
        space = ' '*len(self.__class__.__name__)+'  '
        string = f"{self.__class__.__name__}( A Node called '{self.name}', containing the following top-level objects in its tree:"
        string += "\n"
        for k,v in self._branch.items():
            string += "\n"+space+f"    {k.ljust(24,' ')} \t ({v.__class__.__name__})"
        string += "\n)"
        return string



    ########## .tree methods ###########

    def show_tree(self,root=False):
        """
        Display the object tree. If `root` is False, displays the branch
        of the tree downstream from this node.  If `root` is True, displays
        the full tree from the root node.
        """
        assert(isinstance(root,bool))
        if not root:
            self._branch.print()
        else:
            assert(self.root is not None), "Can't display an unrooted node from its root!"
            self.root._branch.print()

    def add_to_tree(self,node):
        """
        Add an unrooted node as a child of the current, rooted node.
        To move an already rooted node/branch, use `.graft()`.
        To create a rooted node, use `Root()`.
        """
        assert(isinstance(node,Node))
        assert(self.root is not None), "Can't add objects to an unrooted node. See the Node docstring for more info."
        assert(node.root is None), "Can't add a rooted node to a different tree.  Use `.tree(graft=node)` instead."
        node._root = self._root
        self._branch[node.name] = node
        node._treepath = self._treepath+'/'+node.name

    def force_add_to_tree(self,node):
        """
        Add node `node` as a child of the current node, whether or not
        `node` is rooted.  If it's unrooted, performs a simple add. If
        it is rooted, performs a graft, excluding the root metadata
        from `node`.
        """
        try:
            self.add_to_tree(node)
        except AssertionError:
            self.graft(node, merge_metadata=False)


    def get_from_tree(self,name):
        """
        Finds and returns an object from an EMD tree using the string
        key `name`, with '/' delimiters between 'parent/child' nodes.
        Search from the root node by adding a leading '/'; otherwise,
        searches from the current node.
        """
        if name == '':
            return self.root
        elif name[0] != '/':
            return self._branch[name]
        else:
            return self.root._branch[name]

    def _graft(self,node,merge_metadata=True):
        """
        Moves the branch beginning at this node onto another tree at
        targed point `node`.

        This is the reverse of the behavior of self.graft(node), which moves
        the branch on some other tree beginning at `node` onto this tree
        starting at this node. In other words, self.graft() grafts onto self,
        while self._graft() grafts from self.

        Accepts:
            node (Node):
            merge_metadata (True, False, 'copy', or 'overwrite'): if True adds
                the new root's metadata to the old root, skipping entries that
                exist in both; overwrite is identical, but overwrites entries
                that exist in both; if False adds no metadata to the new
                root; if 'copy' adds copies of all metadata from the old root to
                the new root.

        Returns:
            (Node) the root node of the receiving tree
        """
        assert(self.root is not None), "Can't graft an unrooted node; try using .tree(add=node) instead."
        assert(node.root is not None), "Can't graft onto an unrooted node"

        # find upstream and root nodes
        old_root = self.root
        node_list = self._treepath.split('/')
        this_node = node_list.pop()
        treepath = '/'.join(node_list)
        upstream_node = self.get_from_tree(treepath)


        # if grafting from a non-root node
        if this_node != '':

            # remove connection from upstream to this node
            del(upstream_node._branch[this_node])

            # add to a new tree
            self._root = None
            node.add_to_tree(self)

        # if grafting from a root node
        else:

            # add objects one by one
            keys = list(upstream_node._branch.keys())
            for k in keys:
                n = upstream_node.tree(k)
                n._root = None
                node.add_to_tree(n)
                # remove upstream connections
                del(upstream_node._branch[k])


        # add old root metadata to this node
        assert(merge_metadata in [True,False,'copy','overwrite'])
        if merge_metadata is False:
            # add no root metadata
            pass
        elif merge_metadata is True:
            # add old root metadata to new root
            for key in old_root.metadata.keys():
                if key not in node.root.metadata.keys():
                    node.root.metadata = old_root.metadata[key]
        elif merge_metadata == "overwrite":
            # add old root metadata to new root
            for key in old_root.metadata.keys():
                node.root.metadata = old_root.metadata[key]
        else:
            # add copies of old root metadata to new root
            for key in old_root.metadata.keys():
                node.root.metadata = old_root.metadata[key].copy()
        # return
        return node.root


    def graft(self,node,merge_metadata=True):
        """
        Moves the branch beginning node onto this tree at this node.

        For the reverse (i.e. grafting from this tree onto another tree)
        either use that tree's .graft method, or use this tree's ._graft.

        Accepts:
            node (Node):
            merge_metadata (True, False, or 'copy'): if True adds the old root's
                metadata to the new root; if False adds no metadata to the new
                root; if 'copy' adds copies of all metadata from the old root to
                the new root.

        Returns:
            (Node) this tree's root node
        """
        return node._graft(
            self,
            merge_metadata=merge_metadata
        )


    def cut_from_tree(self,root_metadata=True):
        """
        Removes a branch from an object tree at this node.

        A new root node is created under this object
        with this object's name.  Metadata from
        the current root is transferred/not transferred
        to the new root according to the value of `root_metadata`.

        Accepts:
            root_metadata (True, False, or 'copy'): if True adds the old root's
                metadata to the new root; if False adds no metadata to the new
                root; if 'copy' adds copies of all metadata from the old root to
                the new root.

        Returns:
            (Node) the new root node
        """
        new_root = Root( name=self.root.name+'_cut_'+self.name)
        return self._graft(new_root,merge_metadata=root_metadata)


    def tree(
        self,
        arg = None,
        **kwargs,
        ):
        """
        Usages -

            >>> .tree()             # show tree from current node
            >>> .tree(show=True)    # show from root
            >>> .tree(show=False)   # show from current node
            >>> .tree(add=node)     # add a child node
            >>> .tree(get='path')   # return a '/' delimited child node
            >>> .tree(get='/path')  # as above, starting at root
            >>> .tree(cut=True)     # remove/return a branch, keep root metadata
            >>> .tree(cut=False)    # remove/return a branch, discard root md
            >>> .tree(cut='copy')   # remove/return a branch, copy root metadata
            >>> .tree(graft=node)   # remove/graft a branch, keeping root metadata
            >>> .tree(graft=(node,True))    # as above
            >>> .tree(graft=(node,False))   # as above, discard root metadata
            >>> .tree(graft=(node,'copy'))  # as above, copy root metadata

        The show, add, and get methods can be accessed directly with

            >>> .tree(arg)

        for an arg of the appropriate type (bool, Node, and string).
        """

        # if `arg` is passed, choose behavior from its type
        if isinstance(arg,bool):
            self.show_tree(root=arg)
            return
        elif isinstance(arg,Node):
            if 'force' in kwargs:
                self.force_add_to_tree(arg)
            else:
                self.add_to_tree(arg)
            return
        elif isinstance(arg,str):
            return self.get_from_tree(arg)
        else:
            assert(arg is None), f'invalid `arg` type passed to .tree() {type(arg)}'

        # if `arg` is not passed, choose behavior from **kwargs
        if len(kwargs)==0:
            self.show_tree(root=False)
            return
        else:
            assert(len(kwargs)==1),\
                f"kwargs accepts at most 1 argument; recieved {len(kwargs)}"
            k,v = kwargs.popitem()
            assert(k in ['show','add','get','cut','graft']),\
                f"invalid keyword passed to .tree(), {k}"
            if k == 'show':
                assert(isinstance(v,bool)),\
                    f".tree(show=value) requires type(value)==bool, not {type(v)}"
                self.show_tree(root=v)
            elif k == 'add':
                assert(isinstance(v,Node)),\
                    f".tree(add=value) requires `value` to be a Node, received {type(value)}"
                self.add_to_tree(v)
            elif k == 'get':
                assert(isinstance(v,str)),\
                    f".tree(get=value) requires type(value)==str, not {type(v)}"
                return self.get_from_tree(v)
            elif k == 'cut':
                return self.cut_from_tree(root_metadata=v)
            elif k == 'graft':
                if isinstance(v,Node):
                    n,m = v,True
                else:
                    assert(len(v)==2), ".tree(graft=x) requires `x=Node` or `x=(node,metadata)` for some node and `metadata` in (True,False,'copy')"
                    n,m = v
                return self.graft(node=n,merge_metadata=m)
            else:
                raise Exception(f'Invalid arg {k}; must be in (show,add,get,cut,graft)')

    ########## end .tree methods ###########



    # decorator for storing params which generated new data nodes

    @staticmethod
    def newnode(method):
        """
        Decorator which may be added to node methods which product and
        return a new node.  If such a method is decorated with

        >>> @newnode

        then the new node is added to the parent node's tree, and a
        Metadata instance is added to the new node's metadata which
        stores information about how the node was created, namely:
        method's name, the parent's class and name, and all
        the arguments passed to method.
        """

        @functools.wraps(method)
        def wrapper(*args, **kwargs):

            # Get the called method name
            method_name = method.__name__

            # Get the generating class instance and other args
            self,args = args[0],args[1:]

            # Pair args with names
            method_args = inspect.getfullargspec(method)[0]
            d = dict()
            for i in range(len(args)):
                d[method_args[i]] = args[i]

            # Add the generating class type and name
            d['parent_class'] = self.__class__
            d['parent_name'] = self.name
            d['parent_method'] = method_name

            # Combine with kwargs
            kwargs.update(d)
            # and make the metadata
            md = Metadata( name = "_fn_call_" + method_name )
            md._params.update( kwargs )

            # Run the method, get the returned node
            new_node = method(*args,**kwargs)

            # Add the metadata to the node
            new_node.metadata = md

            # Add the new node to the parent node's tree
            self.tree(new_node)

            # Return
            return new_node

        return wrapper




    # read

    # this machinery is designed to work with and be extendible
    # in the context of child classes.

    # to add a new child class, only a new _get_constructor_args()
    # classmethod should be necessary. If a class needs additional
    # construction post-initialization, define a _populate_instance()
    # method. The .from_h5 function should not need any modification.

    @classmethod
    def _get_constructor_args(cls,group):
        """
        Takes a group, and returns a dictionary of args/values
        to pass to the class constructor
        """
        # Nodes contain only metadata
        return {
            'name' : basename(group.name)
        }


    def _populate_instance(self,group):
        """ Nothing to add for Nodes
        """
        pass



    @classmethod
    def from_h5(cls,group):
        """
        Takes an h5py Group which is open in read mode. Confirms that a
        a Node of this name exists in this group, and loads and returns it
        with it's metadata.

        Accepts:
            group (h5py Group)

        Returns:
            (Node)
        """
        # Validate inputs
        er = f"Group {group} is not a valid EMD node"
        assert("emd_group_type" in group.attrs.keys()), er
        assert(group.attrs["emd_group_type"] in EMD_group_types)

        # Make dict of args to build a new generic class instance
        # then make the new class instance
        args = cls._get_constructor_args(group)
        node = cls(**args)

        # some classes needed to be first instantiated, then populated with data
        node._populate_instance(group)

        # Read any metadata
        try:
            grp_metadata = group['metadatabundle']
            for md in grp_metadata.values():
                node.metadata = Metadata.from_h5(md)
                # check for inherited classes
                if md.attrs['python_class'] != 'Metadata':
                    # and attempt promotion
                    try:
                        cls = _get_class(md)
                        inst = cls( name=basename(md.name) )
                        inst._params.update(node.metadata[basename(md.name)]._params)
                        node.metadata = inst
                    except KeyError:
                        print(f"Warning: unable to promote class from Metadata to {md.attrs['python_class']}")
                        pass
        except KeyError:
            pass

        # Return
        return node





    # write
    def to_h5(self,group):
        """
        Takes an h5py Group instance and creates a subgroup containing
        this node, tags indicating the groups EMD type and Python class,
        and any metadata in this node.

        Accepts:
            group (h5py Group)

        Returns:
            (h5py Group) the new node's Group
        """
        # Validate group type
        assert(self.__class__._emd_group_type in EMD_group_types)

        # Make group, add tags
        grp = group.create_group(self.name)
        grp.attrs.create("emd_group_type",self.__class__._emd_group_type)
        grp.attrs.create("python_class",self.__class__.__name__)

        # add metadata
        items = self._metadata.items()
        if len(items)>0:
            # create container group for metadata dictionaries
            grp_metadata = grp.create_group('metadatabundle')
            grp_metadata.attrs.create("emd_group_type","metadatabundle")
            for k,v in items:
                # add each Metadata instance
                self._metadata[k].name = k
                self._metadata[k].to_h5(grp_metadata)

        # return
        return grp









class Root(Node):
    """
    A Node instance with its .root property set to itself.
    """
    _emd_group_type = 'root'

    def __init__(self,name='root'):
        Node.__init__(self,name=name)
        self._treepath = ''
        self._root = self










class Branch:

    def __init__(self):
        self._dict = {}



    # Enables adding items to the Branch's dictionary
    def __setitem__(self, key, value):
        assert(isinstance(value,Node)), f"only Node instances can be added to tree. not type {type(value)}"
        self._dict[key] = value


    # Enables retrieving items at top level,
    # or nested items using 'node1/node2' syntax
    def __getitem__(self,x):
        l = x.split('/')
        try:
            l.remove('')
            l.remove('')
        except ValueError:
            pass
        return self._getitem_from_list(l)

    def _getitem_from_list(self,x):
        if len(x) == 0:
            raise Exception("invalid slice value to tree")

        k = x.pop(0)
        er = f"{k} not found in tree - check keys"
        assert(k in self._dict.keys()), er

        if len(x) == 0:
            return self._dict[k]
        else:
            tree = self._dict[k]._branch
            return tree._getitem_from_list(x)


    def __delitem__(self,x):
        del(self._dict[x])


    # return the top level keys and items
    def keys(self):
        return self._dict.keys()
    def items(self):
        return self._dict.items()



    # print the full tree contents to screen
    def print(self):
        """
        Prints the tree contents to screen.
        """
        print('/')
        self._print_tree_to_screen(self)
        pass

    @staticmethod
    def _print_tree_to_screen(tree, tablevel=0, linelevels=[]):
        """
        """
        if tablevel not in linelevels:
            linelevels.append(tablevel)
        keys = [k for k in tree.keys()]
        N = len(keys)
        for i,k in enumerate(keys):
            string = ''
            string += '|' if 0 in linelevels else ' '
            for idx in range(tablevel):
                l = '|' if idx+1 in linelevels else ' '
                string += '   '+l
            print(string+'---'+k)
            if i == N-1:
                linelevels.remove(tablevel)
            try:
                tree._print_tree_to_screen(
                    tree[k]._branch,
                    tablevel=tablevel+1,
                    linelevels=linelevels)
            except AttributeError:
                pass

        pass



    # Displays the top level objects in the Branch
    def __repr__(self):
        space = ' '*len(self.__class__.__name__)+'  '
        string = f"{self.__class__.__name__}( An object tree containing the following top-level object instances:"
        string += "\n"
        for k,v in self._dict.items():
            string += "\n"+space+f"    {k} \t\t ({v.__class__.__name__})"
        string += "\n)"
        return string







# Defines the Array class, which stores any N-dimensional array-like data.
# Implements the EMD file standard - https://emdatasets.com/format

from typing import Optional,Union
import numpy as np
import h5py
from numbers import Number
from os.path import basename

from emdfile.classes.tree import Node

class Array(Node):
    """
    A class which stores any N-dimensional array-like data, plus basic metadata:
    a name and units, as well as calibrations for each axis of the array, and names
    and units for those axis calibrations.

    In the simplest usage, only a data array is passed:

    >>> ar = Array(np.ones((20,20,256,256)))

    will create an array instance whose data is the numpy array passed, and with
    automatically populated dimension calibrations in units of pixels.

    Additional arguments may be passed to populate the object metadata:

    >>> ar = Array(
    >>>     np.ones((20,20,256,256)),
    >>>     name = 'test_array',
    >>>     units = 'intensity',
    >>>     dims = [
    >>>         [0,5],
    >>>         [0,5],
    >>>         [0,0.01],
    >>>         [0,0.01]
    >>>     ],
    >>>     dim_units = [
    >>>         'nm',
    >>>         'nm',
    >>>         'A^-1',
    >>>         'A^-1'
    >>>     ],
    >>>     dim_names = [
    >>>         'rx',
    >>>         'ry',
    >>>         'qx',
    >>>         'qy'
    >>>     ],
    >>> )

    will create an array with a name and units for its data, where its first two
    dimensions are in units of nanometers, have pixel sizes of 5nm, and are
    described by the handles 'rx' and 'ry', and where its last two dimensions
    are in units of inverse Angstroms, have pixels sizes of 0.01A^-1, and are
    described by the handles 'qx' and 'qy'.

    Arrays in which the length of each pixel is non-constant are also
    supported.  For instance,

    >>> x = np.logspace(0,1,100)
    >>> y = np.sin(x)
    >>> ar = Array(
    >>>     y,
    >>>     dims = [
    >>>         x
    >>>     ]
    >>> )

    generates an array representing the values of the sine function sampled
    100 times along a logarithmic interval from 1 to 10. In this example,
    this data could then be plotted with, e.g.

    >>> plt.scatter(ar.dims[0], ar.data)

    If the `slicelabels` keyword is passed, the first N-1 dimensions of the
    array are treated normally, while the final dimension is used to represent
    distinct arrays which share a common shape and set of dim vectors.  Thus

    >>> ar = Array(
    >>>     np.ones((50,50,4)),
    >>>     name = 'test_array_stack',
    >>>     units = 'intensity',
    >>>     dims = [
    >>>         [0,2],
    >>>         [0,2]
    >>>     ],
    >>>     dim_units = [
    >>>         'nm',
    >>>         'nm'
    >>>     ],
    >>>     dim_names = [
    >>>         'rx',
    >>>         'ry'
    >>>     ],
    >>>     slicelabels = [
    >>>         'a',
    >>>         'b',
    >>>         'c',
    >>>         'd'
    >>>     ]
    >>> )

    will generate a single Array instance containing 4 arrays which each have
    a shape (50,50) and a common set of dim vectors ['rx','ry'], and which
    can be indexed into with the names assigned in `slicelabels` using

    >>> ar.get_slice('a')

    which will return a 2D (non-stack-like) Array instance with shape (50,50)
    and the dims assigned above.  The Array attribute .rank is equal to the
    number of dimensions for a non-stack-like Array, and is equal to N-1
    for stack-like arrays.

    """
    _emd_group_type = 'array'

    def __init__(
        self,
        data: np.ndarray,
        name: Optional[str] = 'array',
        units: Optional[str] = '',
        dims: Optional[list] = None,
        dim_names: Optional[list] = None,
        dim_units: Optional[list] = None,
        slicelabels = None
        ):
        """
        Accepts:
            data (np.ndarray): the data
            name (str): the name of the Array
            units (str): units for the pixel values
            dims (variable): calibration vectors for each of the axes of the data
                array.  Valid values for each element of the list are None,
                a number, a 2-element list/array, or an M-element list/array
                where M is the data array.  If None is passed, the dim will be
                populated with integer values starting at 0 and its units will
                be set to pixels.  If a number is passed, the dim is populated
                with a vector beginning at zero and increasing linearly by this
                step size.  If a 2-element list/array is passed, the dim is
                populated with a linear vector with these two numbers as the first
                two elements.  If a list/array of length M is passed, this is used
                as the dim vector, (and must therefore match this dimension's
                length). If dims recieves a list of fewer than N arguments for an
                N-dimensional data array, the extra dimensions are populated as if
                None were passed, using integer pixel values. If the `dims`
                parameter is not passed, all dim vectors are populated this way.
            dim_units (list): the units for the calibration dim vectors. If
                nothing is passed, dims vectors which have been populated
                automatically with integers corresponding to pixel numbers
                will be assigned units of 'pixels', and any other dim vectors
                will be assigned units of 'unknown'.  If a list with length <
                the array dimensions, the passed values are assumed to apply
                to the first N dimensions, and the remaining values are
                populated with 'pixels' or 'unknown' as above.
            dim_names (list): labels for each axis of the data array. Values
                which are not passed, following the same logic as described
                above, will be autopopulated with the name "dim#" where #
                is the axis number.
            slicelabels (None or True or list): if not None, must be True or a
                list of strings, indicating a "stack-like" array.  In this case,
                the first N-1 dimensions of the array are treated normally, in
                the sense of populating dims, dim_names, and dim_units, while the
                final dimension is treated distinctly: it indexes into
                distinct arrays which share a set of dimension attributes, and
                can be sliced into using the string labels from the `slicelabels`
                list, with the syntax array['label'] or array.get_slice('label').
                If `slicelabels` is `True` or is a list with length less than the
                final dimension length, unassigned dimensions are autopopulated
                with labels `array{i}`. The flag array.is_stack is set to True
                and the array.rank attribute is set to N-1.

        Returns:
            A new Array instance
        """
        super().__init__()

        # populate data and metadata
        self.data = data
        self.name = name
        self.units = units

        ## For array stacks,
        ## setup labels and N+1'th dimension
        if slicelabels is None:
            self.is_stack = False
            self.slicelabels = None
        else:
            self.is_stack = True
            # Populate labels
            if slicelabels is True:
                slicelabels = [f'array{i}' for i in range(self.depth)]
            elif len(slicelabels) < self.depth:
                slicelabels = np.concatenate((slicelabels,
                    [f'array{i}' for i in range(len(slicelabels),self.depth)]))
            else:
                slicelabels = slicelabels[:self.depth]
            self.slicelabels = Labels(slicelabels)


        ## Setup dim vectors

        # set initial state
        self._dims = tuple([None for i in range(self.rank)])
        self._dim_units = tuple(['unknown' for i in range(self.rank)])
        self._dim_names = tuple([f"dim{i}" for i in range(self.rank)])

        # expand dims, dim_units, and dim_names to lists of length = rank,
        # padding with None if the lists are too short

        # setup dims
        if dims is None:
            dims = [None for i in range(self.rank)]
        else:
            assert(isinstance(dims,(list,tuple))), f"dims must be None or a list or tuple, not type {type(dims)}"
            if len(dims) < (self.rank):
                dims = list(dims) + [None for i in range(self.rank-len(dims))]
            else:
                dims = dims[:self.rank]
        # set flag for if units are pixels
        dim_in_pixels = np.zeros(self.rank, dtype=bool)
        for idx in range(self.rank):
            dim_in_pixels[idx] = dims[idx] is None

        if dim_units is None:
            dim_units = ['unknown' for i in range(self.rank)]
        else:
            assert(isinstance(dim_units,(list,tuple))), f"dim_units must be None or a list or tuple, not type {type(dim_units)}"
            if len(dim_units) < (self.rank):
                dim_units = list(dim_units) + ['unknown' for i in range(rank-len(dim_units))]
            else:
                dim_units = dim_units[:self.rank]
        dim_units = np.array(dim_units)
        dim_units[dim_in_pixels] = 'pixels'
        dim_units = tuple(dim_units)

        if dim_names is None:
            dim_names = [f"dim{i}" for i in range(self.rank)]
        else:
            assert(isinstance(dim_names,(list,tuple))), f"dim_names must be None or a list or tuple, not type {type(dim_names)}"
            if len(dim_names) < (self.rank):
                dim_names = list(dim_names) + [f"dim{i+len(dim_names)}" for i in range(rank-len(dim_names))]
            else:
                dim_names = dim_names[:self.rank]

        # set the dims
        for idx,(d,du,dn) in enumerate(zip(dims,dim_units,dim_names)):
            self.set_dim(
                idx,
                dim = d,
                units = du,
                name = dn
            )



    # dim vector setter/getter properties and methods

    @property
    def dims(self):
        return self._dims

    def get_dim(self,n):
        """ Return the n'th dim vector
        """
        assert(isinstance(n,(int,np.integer))), f"Can't retrieve the {n}th dim vector - {n} must be an integer, not type {type(n)}."
        assert(n < self.rank), f"Can't retrieve the {n}th dim vector - {n} must be < {self.rank}"
        return self.dims[n]

    # alias
    dim = get_dim

    def set_dim(
        self,
        n:int,
        dim:Union[list,np.ndarray],
        units:Optional[str]=None,
        name:Optional[str]=None
        ):
        """
        Sets the n'th dim vector, using `dim` as described in the Array
        documentation. If `units` and/or `name` are passed, sets these
        values for the n'th dim vector.

        Accepts:
            n (int): specifies which dim vector
            dim (list or array): length must be either 2, or equal to the
                length of the n'th axis of the data array
            units (Optional, str):
            name: (Optional, str):
        """
        assert(isinstance(n,(int,np.integer))), f"Can't set the {n}th dim vector - {n} must be an integer, not type {type(n)}."
        assert(n < self.rank), f"Can't set the {n}th dim vector - {n} must be < {self.rank}"
        length = self.shape[n]
        # unpack dim, if a number or len 2 list were passed
        new_dim = self._unpack_dim(dim,length)
        # set new dim
        self._dims = list(self._dims)
        self._dims[n] = new_dim
        self._dims = tuple(self._dims)
        if units is not None:
            self.set_dim_units(n,units)
        if name is not None:
            self.set_dim_name(n,name)

    @property
    def dim_units(self):
        return self._dim_units

    def get_dim_units(self,n):
        """ Return the n'th dim vector units
        """
        assert(isinstance(n,(int,np.integer))), f"Can't retrieve the {n}th dim vector - {n} must be an integer, not type {type(n)}."
        assert(n < self.rank), f"Can't retrieve the {n}th dim vector - {n} must be < {self.rank}"
        return self.dim_units[n]

    def set_dim_units(
        self,
        n:int,
        units:str,
        ):
        """
        Sets the n'th dim vector units to `units`.

        Accepts:
            n (int): specifies which dim vector
            units (str): new units
        """
        assert(isinstance(n,(int,np.integer))), f"Can't set the {n}th dim vector - {n} must be an integer, not type {type(n)}."
        assert(n < self.rank), f"Can't set the {n}th dim vector - {n} must be < {self.rank}"
        self._dim_units = list(self._dim_units)
        self._dim_units[n] = units
        self._dim_units = tuple(self._dim_units)

    @property
    def dim_names(self):
        return self._dim_names

    def get_dim_name(self,n):
        """ Get the n'th dim vector name
        """
        assert(isinstance(n,(int,np.integer))), f"Can't retrieve the {n}th dim vector - {n} must be an integer, not type {type(n)}."
        assert(n < self.rank), f"Can't retrieve the {n}th dim vector - {n} must be < {self.rank}"
        return self.dim_names[n]

    def set_dim_name(
        self,
        n:int,
        name:str,
        ):
        """
        Sets the n'th dim vector name to `name`.

        Accepts:
            n (int): specifies which dim vector
            name (str): new name
        """
        assert(isinstance(n,(int,np.integer))), f"Can't set the {n}th dim vector - {n} must be an integer, not type {type(n)}."
        assert(n < self.rank), f"Can't set the {n}th dim vector - {n} must be < {self.rank}"
        self._dim_names = list(self._dim_names)
        self._dim_names[n] = name
        self._dim_names = tuple(self._dim_names)


    @staticmethod
    def _unpack_dim(dim,length):
        """
        Given a dim vector as passed at instantiation and the expected length
        of this dimension of the array, this function checks the passed dim
        vector length, and checks the dim vector type.  For number-like dim-
        vectors:
            -if it is a number, turns it into the list [0,number] and proceeds
                as below
            -if it has length 2, linearly extends the vector to its full length
            -if it has length `length`, returns the vector as is
            -if it has any other length, raises an Exception.

        For string-like dim vectors, the length must match the array dimension
        length.

        Accepts:
            dim (list or array)
            length (int)

        Returns
            the unpacked dim vector
        """
        # Expand None
        if dim is None:
            dim = 1

        # Expand single numbers
        if isinstance(dim,Number):
            dim = [0,dim]

        N = len(dim)

        # for string dimensions (used for stack arrays):
        if not isinstance(dim[0],Number):
            assert(N == length), f"For non-numerical dims, the dim vector length must match the array dimension length. Recieved a dim vector of length {N} for an array dimension length of {length}."

        # For number-like dimensions:
        if N == length:
            return dim
        elif N == 2:
            start,step = dim[0],dim[1]-dim[0]
            stop = start + step*length
            return np.arange(start,stop,step)
        else:
            raise Exception(f"dim vector length must be either 2 or equal to the length of the corresponding array dimension; dim vector length was {dim} and the array dimension length was {length}")

    def _dim_is_linear(self,dim,length):
        """
        Returns True if a dim is linear, else returns False
        """
        try:
            dim_expanded = self._unpack_dim(dim[:2],length)
            return np.array_equal(dim,dim_expanded)
        except IndexError:
            return True





    # Shape properties

    @property
    def shape(self):
        if not self.is_stack:
            return self.data.shape
        else:
            return self.data.shape[1:]

    @property
    def depth(self):
        if not self.is_stack:
            return 0
        else:
            return self.data.shape[0]

    @property
    def rank(self):
        if not self.is_stack:
            return self.data.ndim
        else:
            return self.data.ndim - 1


    ## Slicing

    def get_slice(self,label,name=None):
        idx = self.slicelabels._dict[label]
        return Array(
            data = self.data[idx],
            name = name if name is not None else self.name+'_'+label,
            units = self.units,
            dims = self.dims,
            dim_units = self.dim_units,
            dim_names = self.dim_names
        )

    def __getitem__(self,x):
        if isinstance(x,str):
            return self.get_slice(x)
        elif isinstance(x,tuple) and isinstance(x[0],str):
            return self.get_slice(x[0])[x[1:]]
        else:
            return self.data[x]






    ## Representation to standard output

    def __repr__(self):

        if not self.is_stack:
            space = ' '*len(self.__class__.__name__)+'  '
            string = f"{self.__class__.__name__}( A {self.rank}-dimensional array of shape {self.shape} called '{self.name}',"
            string += "\n"+space+"with dimensions:"
            string += "\n"

        else:
            space = ' '*len(self.__class__.__name__)+'  '
            string = f"{self.__class__.__name__}( A stack of {self.depth} Arrays with {self.rank}-dimensions and shape {self.shape}, called '{self.name}'"
            string += "\n"
            string += "\n" +space + "The labels are:"
            for label in self.slicelabels:
                string += "\n" + space + f"    {label}"
            string += "\n"
            string += "\n"
            string += "\n" + space + "The Array dimensions are:"

        for n in range(self.rank):
            if len(self.dims[n]) >= 3:
                string += "\n"+space+f"    {self.dim_names[n]} = [{self.dims[n][0]},{self.dims[n][1]},{self.dims[n][2]},...] {self.dim_units[n]}"
            if len(self.dims[n]) == 2:
                string += "\n"+space+f"    {self.dim_names[n]} = [{self.dims[n][0]},{self.dims[n][1]}] {self.dim_units[n]}"
            elif len(self.dims[n]) == 1:
                string += "\n"+space+f"    {self.dim_names[n]} = [{self.dims[n][0]}] {self.dim_units[n]}"
            else:
                string += "\n"+space+f"    {self.dim_names[n]} = [] {self.dim_units[n]}"
            if not self._dim_is_linear(self.dims[n],self.shape[n]):
                string += "  (*non-linear*)"
        string += "\n)"


        return string



    # HDF5 read/write

    # write
    def to_h5(self,group):
        """
        Takes an h5py Group instance and creates a subgroup containing
        this Array, tags indicating its EMD type and Python class,
        and the array's data and metadata.

        Accepts:
            group (h5py Group)

        Returns:
            (h5py Group) the new array's Group
        """
        # Construct group and add metadata
        grp = Node.to_h5(self,group)

        # add the data
        data = grp.create_dataset(
            "data",
            shape = self.data.shape,
            data = self.data
            #dtype = type(self.data)
        )
        data.attrs.create('units',self.units) # save 'units' but not 'name' - 'name' is the group name

        # Add the normal dim vectors
        for n in range(self.rank):

            # unpack info
            dim = self.dims[n]
            name = self.dim_names[n]
            units = self.dim_units[n]

            # compress the dim vector if it's linear
            if self._dim_is_linear(dim,self.shape[n]):
                dim = dim[:2]

            # write
            dset = grp.create_dataset(
                f"dim{n}",
                data = dim
            )

            dset.attrs.create('name',str(name))
            dset.attrs.create('units',str(units))

        # Add stack dim vector, if present
        if self.is_stack:
            n = self.rank
            dim = [s.encode('utf-8') for s in self.slicelabels]

            # write
            dset = grp.create_dataset(
                f"dim{n}",
                data = dim
            )
            dset.attrs.create('name','_labels_')

        # return
        return grp



    # read
    @classmethod
    def _get_constructor_args(cls,group):
        """
        Returns a dictionary of args/values to pass to the class constructor
        """
        # get data
        dset = group['data']
        data = dset[:]
        units = dset.attrs['units']
        rank = len(data.shape)

        # determine if this is a stack array
        last_dim = group[f"dim{rank-1}"]
        if last_dim.attrs['name'] == '_labels_':
            is_stack = True
            normal_dims = rank-1
        else:
            is_stack = False
            normal_dims = rank

        # get dim vectors
        dims = []
        dim_units = []
        dim_names = []
        for n in range(normal_dims):
            dim_dset = group[f"dim{n}"]
            dims.append(dim_dset[:])
            dim_units.append(dim_dset.attrs['units'])
            dim_names.append(dim_dset.attrs['name'])

        # if it's a stack array, get the labels
        if is_stack:
            slicelabels = last_dim[:]
            slicelabels = [s.decode('utf-8') for s in slicelabels]
        else:
            slicelabels = None

        # make args dictionary and return
        return {
            'data' : data,
            'name' : basename(group.name),
            'units' : units,
            'dims' : dims,
            'dim_names' : dim_names,
            'dim_units' : dim_units,
            'slicelabels' : slicelabels
        }





########### END OF CLASS ###########


# List subclass for accessing data slices with a dict
class Labels(list):

    def __init__(self,x=[]):
        list.__init__(self,x)
        self.setup_labels_dict()

    def __setitem__(self,idx,label):
        label_old = self[idx]
        del(self._dict[label_old])
        list.__setitem__(self,idx,label)
        self._dict[label] = idx

    def setup_labels_dict(self):
        self._dict = {}
        for idx,label in enumerate(self):
            self._dict[label] = idx





from emdfile import PointListArray, Metadata, _read_metadata
import numpy as np

class PointListArraySubclass(PointListArray):
    """
    """

    def __init__(
        self,
        shape,
        name = 'my_pointlistarray_subclass',
        ):
        """
        A 2D ragged array of float32's. Shape must be a 2-tuple.
        Initializes empty.
        """

        # initialize
        PointListArray.__init__(
            self,
            name = name,
            dtype = np.float32,
            shape = shape
        )

        pass



    # Read methods
    @classmethod
    def _get_constructor_args(cls,group):
        """
        Returns a dictionary of arguments to pass to __init__
        """
        # get array constructor args
        kwargs = PointListArray._get_constructor_args(group)

        # make the dictionary of constructor arguments and return
        constructor_args = {
            'shape' : kwargs['shape'],
            'name' : kwargs['name']
        }
        return constructor_args


    # PointListArray already has a _populate instance method which
    # will add the ragged data array.







import emdfile as emd
import numpy as np


def test_array_repr():

    # multidim array
    s = (4,5,6,7)
    ar1 = emd.Array(
        name = 'ar1',
        data = np.arange(np.prod(s)).reshape(s)
    )

    # multidim array with len-2 dim
    s = (2,5,6,7)
    ar2 = emd.Array(
        name = 'ar2',
        data = np.arange(np.prod(s)).reshape(s)
    )

    # multidim array with len-2 dim
    s = (4,2,6,7)
    ar3 = emd.Array(
        name = 'ar3',
        data = np.arange(np.prod(s)).reshape(s)
    )

    # multidim array with len-1 dim
    s = (1,5,6,7)
    ar4 = emd.Array(
        name = 'ar4',
        data = np.arange(np.prod(s)).reshape(s)
    )

    # multidim array with len-1 dim
    s = (4,1,6,7)
    ar5 = emd.Array(
        name = 'ar5',
        data = np.arange(np.prod(s)).reshape(s)
    )

    # single dim array
    ar6 = emd.Array(
        name = 'ar6',
        data = np.arange(10)
    )

    # single dim array, len 2
    ar6 = emd.Array(
        name = 'ar6',
        data = np.arange(2)
    )

    # single dim array, len 1
    ar7 = emd.Array(
        name = 'ar7',
        data = np.arange(1)
    )

    # single dim array, len 0
    ar8 = emd.Array(
        name = 'ar8',
        data = np.arange(0)
    )

    # test __repr__
    print(ar1)
    print(ar2)
    print(ar3)
    print(ar4)
    print(ar5)
    print(ar6)
    print(ar7)
    print(ar8)





import numpy as np
from numbers import Number
from typing import Optional
from os.path import basename



class Metadata:
    """
    Stores metadata in the form of a flat (non-nested) dictionary.
    Keys are arbitrary strings.  Values may be strings, numbers, arrays,
    or lists of the above types.

    Usage:

        >>> meta = Metadata()
        >>> meta['param'] = value
        >>> val = meta['param']

    If the parameter has not been set, the getter methods return None.
    """
    _emd_group_type = 'metadata'

    def __init__(
        self,
        name: Optional[str] = 'metadata',
        data: Optional[dict] = None
        ):
        """
         Args:
            name (Optional, string):
        """
        self.name = name
        self._params = {}

        if data is not None:
            assert(isinstance(data,dict)), f"`data` must be a dict, not type {type(data)}"
            self._params.update(data)


    ### __get/setitem__

    def __getitem__(self,x):
        return self._params[x]
    def __setitem__(self,k,v):
        self._params[k] = v


    @property
    def keys(self):
        return self._params.keys()


    def copy(self,name=None):
        """
        """
        if name is None: name = self.name+"_copy"
        md = Metadata(name=name)
        md._params.update(self._params)
        return md



    def __repr__(self):

        space = ' '*len(self.__class__.__name__)+'  '
        string = f"{self.__class__.__name__}( A Metadata instance called '{self.name}', containing the following fields:"
        string += "\n"

        maxlen = 0
        for k in self._params.keys():
            if len(k)>maxlen: maxlen=len(k)

        for k,v in self._params.items():
            if isinstance(v,np.ndarray):
                v = f"{v.ndim}D-array"
            string += "\n"+space+f"{k}:{(maxlen-len(k)+3)*' '}{str(v)}"
        string += "\n)"

        return string



    # HDF5 i/o

    # write
    def to_h5(self,group):
        """
        Accepts an h5py Group which is open in write or append mode. Writes
        a new group with this object's name and saves its metadata in it.

        Accepts:
            group (h5py Group)
        """
        # Make a new group
        grp = group.create_group(self.name)
        grp.attrs.create("emd_group_type","metadata")
        grp.attrs.create("python_class",self.__class__.__name__)

        # Loop + save
        for k,v in self._params.items():
            self._save_item(k,v,grp)


    def _save_item(self,k,v,grp):
        """
        For some (key, value, group), saves the piece of metadata to group.
        """
        dict
        if isinstance(v,dict):
            _grp = grp.create_group(k)
            _grp.attrs['type'] = 'dict'
            for _k,_v in v.items():
                self._save_item(_k,_v,_grp)

        # None
        elif v is None:
            v = "_None"
            v = v.encode('utf-8')  # convert to byte string
            dset = grp.create_dataset(k, data=v)
            dset.attrs['type'] = 'None'.encode('utf-8')

        # strings
        elif isinstance(v, str):
            v = v.encode('utf-8')  # convert to byte string
            dset = grp.create_dataset(k, data=v)
            dset.attrs['type'] = 'string'.encode('utf-8')

        # bools
        elif isinstance(v, bool):
            dset = grp.create_dataset(k, data=v, dtype=bool)
            dset.attrs['type'] = 'bool'.encode('utf-8')

        # numbers
        elif isinstance(v, Number):
            dset = grp.create_dataset(k, data=v, dtype=type(v))
            dset.attrs['type'] = 'number'.encode('utf-8')

        # arrays
        elif isinstance(v, np.ndarray):
            dset = grp.create_dataset(k, data=v, dtype=v.dtype)
            dset.attrs['type'] = 'array'.encode('utf-8')

        # tuples
        elif isinstance(v, tuple):

            # empty
            if len(v) == 0:
                dset = grp.create_dataset(k, data=v)
                dset.attrs['type'] = 'tuple'.encode('utf-8')

            # of numbers
            elif isinstance(v[0], Number):
                dset = grp.create_dataset(k, data=v)
                dset.attrs['type'] = 'tuple'.encode('utf-8')

            # of tuples
            elif any([isinstance(v[i], tuple) for i in range(len(v))]):
                dset_grp = grp.create_group(k)
                dset_grp.attrs['type'] = 'tuple_of_tuples'.encode('utf-8')
                dset_grp.attrs['length'] = len(v)
                for i,x in enumerate(v):
                    dset_grp.create_dataset(
                        str(i),
                        data=x)

            # of arrays
            elif isinstance(v[0], np.ndarray):
                dset_grp = grp.create_group(k)
                dset_grp.attrs['type'] = 'tuple_of_arrays'.encode('utf-8')
                dset_grp.attrs['length'] = len(v)
                for i,ar in enumerate(v):
                    dset_grp.create_dataset(
                        str(i),
                        data=ar,
                        dtype=ar.dtype)

            # of strings
            elif isinstance(v[0], str):
                dset_grp = grp.create_group(k)
                dset_grp.attrs['type'] = 'tuple_of_strings'.encode('utf-8')
                dset_grp.attrs['length'] = len(v)
                for i,s in enumerate(v):
                    dset_grp.create_dataset(
                        str(i),
                        data=s.encode('utf-8'))

            else:
                er = f"Metadata only supports writing tuples with numeric and array-like arguments; found type {type(v[0])}"
                raise Exception(er)

        # lists
        elif isinstance(v, list):

            # empty
            if len(v) == 0:
                dset = grp.create_dataset(k, data=v)
                dset.attrs['type'] = 'list'.encode('utf-8')

            # of numbers
            elif isinstance(v[0], Number):
                dset = grp.create_dataset(k, data=v)
                dset.attrs['type'] = 'list'.encode('utf-8')

            # of arrays
            elif isinstance(v[0], np.ndarray):
                dset_grp = grp.create_group(k)
                dset_grp.attrs['type'] = 'list_of_arrays'.encode('utf-8')
                dset_grp.attrs['length'] = len(v)
                for i,ar in enumerate(v):
                    dset_grp.create_dataset(
                        str(i),
                        data=ar,
                        dtype=ar.dtype)

            # of strings
            elif isinstance(v[0], str):
                dset_grp = grp.create_group(k)
                dset_grp.attrs['type'] = 'list_of_strings'.encode('utf-8')
                dset_grp.attrs['length'] = len(v)
                for i,s in enumerate(v):
                    dset_grp.create_dataset(
                        str(i),
                        data=s.encode('utf-8'))

            else:
                er = f"Metadata only supports writing lists with numeric and array-like arguments; found type {type(v[0])}"
                raise Exception(er)

        else:
            er = f"Metadata supports writing numbers, bools, strings, arrays, tuples of numbers or arrays, and lists of numbers or arrays. Found an unsupported type {type(v[0])}"
            raise Exception(er)


    # read
    @classmethod
    def from_h5(cls,group):
        """
        Accepts an h5py Group which is open in read mode, confirms that
        it represents an EMD MetadataDict group, then loads and returns it
        as a Metadata instance.

        Accepts:
            group (HDF5 group)

        Returns:
            (Metadata)
        """
        # Validate inputs
        er = f"Group {group} is not a valid EMD Metadata group"
        assert("emd_group_type" in group.attrs.keys()), er
        assert(group.attrs["emd_group_type"] == "metadata"), er

        # Loop and get data
        data = {}
        cls._read_items(group,data)

        # make Metadata instance, add data, and return
        md = cls(basename(group.name))
        md._params.update(data)
        return md


    @classmethod
    def _read_items(cls, group, data):
        """
        For some (group,dict), reads the metadata from group, adding
        items into dict. Nothing is returned.
        """
        for k,v in group.items():
            data[k] = cls._read_item(k,v,group)

    @classmethod
    def _read_item(cls, k, v, group):
        """
        For some (key,value,group), reads and returns the piece of metadata
        from group.
        """
        # Get type
        try:
            t = group[k].attrs['type']
            if isinstance(t,bytes):
                t = t.decode('utf-8')
        except KeyError:
            raise Exception(f"unrecognized Metadata value type {type(v)}")

        # Read by type...

        # dict
        if t == 'dict':
            data = {}
            grp = group[k]
            cls._read_items(grp,data)
            v = data

        # None
        elif t == 'None':
            v = None

        # strings
        elif t == 'string':
            v = v[...].item()
            v = v.decode('utf-8')
            v = v if v != "_None" else None

        # numbers
        elif t == 'number':
            v = v[...].item()

        # bools
        elif t == 'bool':
            v = v[...].item()

        # array
        elif t == 'array':
            v = np.array(v)

        # tuples of numbers
        elif t == 'tuple':
            v = tuple(v[...])

        # tuples of arrays
        elif t == 'tuple_of_arrays':
            L = group[k].attrs['length']
            tup = []
            for l in range(L):
                tup.append(np.array(v[str(l)]))
            v = tuple(tup)

        # tuples of tuples
        elif t == 'tuple_of_tuples':
            L = group[k].attrs['length']
            tup = []
            for l in range(L):
                x = v[str(l)][...]
                if x.ndim == 0:
                    x = x.item()
                else:
                    x = tuple(x)
                tup.append(x)
            v = tuple(tup)

        # tuples of strings
        elif t == 'tuple_of_strings':
            L = group[k].attrs['length']
            tup = []
            for l in range(L):
                s = v[str(l)][...].item().decode('utf-8')
                tup.append(s)
            v = tuple(tup)

        # lists of numbers
        elif t == 'list':
            v = list(v[...])

        # lists of arrays
        elif t == 'list_of_arrays':
            L = group[k].attrs['length']
            _list = []
            for l in range(L):
                _list.append(np.array(v[str(l)]))
            v = _list

        # list of strings
        elif t == 'list_of_strings':
            L = group[k].attrs['length']
            _list = []
            for l in range(L):
                s = v[str(l)][...].item().decode('utf-8')
                _list.append(s)
            v = _list

        else:
            raise Exception(f"unrecognized Metadata value type {t}")

        # return
        return v






import numpy as np
from os.path import join,exists
from os import remove
from numpy import array_equal
import h5py

from emdfile import _TESTPATH
from emdfile import save, read
from emdfile.read import _is_EMD_file,_get_EMD_rootgroups
from emdfile.classes import (
    Node,
    Root,
    Metadata,
    Array,
    PointList,
    PointListArray
)

# Set paths
dirpath = _TESTPATH
single_node_path = join(dirpath,"test_h5_single_node.h5")
unrooted_node_path = join(dirpath,"test_h5_unrooted_node.h5")
whole_tree_path = join(dirpath,"test_h5_whole_tree.h5")
subtree_path = join(dirpath,"test_h5_subtree.h5")
subtree_noroot_path = join(dirpath,"test_h5_subtree_noroot.h5")
subtree_append_to_path = join(dirpath,"test_h5_append_to.h5")
subtree_append_over_path = join(dirpath,"test_h5_append_over.h5")


class TreeBuilder():

    def _build_tree(self):
        """
        Helper function which builds the following tree of unique objects.

        Object Class          Name

        Root                  'root'
          |-Metadata          'metadata_root'
          |-Metadata          'metadata_root'
          |-PointList         'pointlist3'
          | |-PointList       'pointlist4'
          |   |-Metadata      'metadata3'
          |-Array             'array
            |-Metadata        'metadata'
            |-Metadata        'metadata2'
            |-PointListArray  'pointlistarray'
              |-PointList     'pointlist'
              |-PointList     'pointlist2'

        Also creates a single unrooted Array instance called 'unrooted_array'
        """
        # make some objects
        ar = Array(
            data = np.array([[1,2],[3,4]])
        )
        ar2 = Array(
            data = np.array([[10,20,30],[40,50,60]]),
            name = 'array'
        )
        ar_unrooted = Array(
            data = np.array([[5,6],[7,8]]),
            name = 'unrooted_array'
        )
        ar_unrooted.set_dim(
            0,
            dim = 0.25,
            units = 'ponies',
            name = 'pony_axis'
        )
        ar_unrooted.set_dim( 1, dim = [0,1.5] )
        ar_unrooted.set_dim_units(1,'cows')
        pl = PointList(
            data = np.zeros(
                5,
                dtype = [
                    ('x',float),
                    ('y',float)
                ]
            )
        )
        pl2 = PointList(
            data = np.ones(
                4,
                dtype = [
                    ('qx',float),
                    ('qy',float)
                ]
            ),
            name = 'pointlist2'
        )
        pl3 = PointList(
            data = np.ones(
                8,
                dtype = [
                    ('r',float),
                    ('theta',float)
                ]
            ),
            name = 'pointlist3'
        )
        pl4 = PointList(
            data = np.ones(
                7,
                dtype = [
                    ('z',float),
                    ('c',float)
                ]
            ),
            name = 'pointlist4'
        )
        pl5 = PointList(
            data = np.ones(
                6,
                dtype = [
                    ('l',float),
                    ('m',float)
                ]
            ),
            name = 'pointlist5'
        )
        pla = PointListArray(
            shape = (3,4),
            dtype = [
                ('a',float),
                ('b',float)
            ]
        )
        md = Metadata()
        md._params.update({
            'duck':'steve'
        })
        md2 = Metadata(name='metadata2')
        md2._params.update({
            'duck':'steven'
        })
        md3 = Metadata(name='metadata3')
        md3._params.update({
            'duck':'stevo'
        })
        md_root = Metadata(name='metadata_root')
        md_root._params.update({
            'cow':'ben'
        })
        md_root2 = Metadata(name='metadata_root2')
        md_root2._params.update({
            'cow':'benjamin'
        })
        root = Root()
        root2 = Root()

        # construct tree
        root.metadata = md_root
        root.metadata = md_root2
        root.tree(pl3)
        pl3.tree(pl4)
        pl4.metadata = md3
        root.tree(ar)
        ar.metadata = md
        ar.metadata = md2
        ar.tree(pla)
        pla.tree(pl)
        pla.tree(pl2)

        # construct smaller tree
        root2.tree(ar2)
        ar2.tree(pl5)

        # assign data instances to variables in the class instance scope
        self.root = root
        self.md = md
        self.md2 = md2
        self.md3 = md3
        self.md_root = md_root
        self.md_root2 = md_root2
        self.ar = ar
        self.ar2 = ar2
        self.ar_unrooted = ar_unrooted
        self.pl = pl
        self.pl2 = pl2
        self.pl3 = pl3
        self.pl4 = pl4
        self.pl5 = pl5
        self.pla = pla

        return

    def _del_tree(self):
        del(self.root)
        del(self.md)
        del(self.md2)
        del(self.md3)
        del(self.md_root)
        del(self.md_root2)
        del(self.ar)
        del(self.ar_unrooted)
        del(self.pl)
        del(self.pl2)
        del(self.pl3)
        del(self.pl4)
        del(self.pla)







class TestTree(TreeBuilder):

    def setup_method(self, method):
        self._build_tree()

    def teardown_method(self, method):
        self._del_tree()


    @staticmethod
    def pointlist_equal(pl1,pl2):
        """
        Return true iff pl1 and pl2 are equal
        """
        # allow for pl1 and pl2 to have the same fields in different order
        k1,k2 = pl1.fields,pl2.fields
        assert(all([k in k1 for k in k2]))
        assert(all([k in k2 for k in k1]))
        for k in k1:
            try:
                assert(array_equal(pl1.data[k],pl2.data[k]))
            except AssertionError:
                return False
        return True


    def test_nodes(self):
        """
        Confirm that nodes are Nodes, metadata are not nodes
        """
        # confirm that EMD objects >0 are Node instances, =0 are not
        assert(isinstance(self.ar,Node))
        assert(isinstance(self.pl,Node))
        assert(isinstance(self.pla,Node))
        assert(not(isinstance(self.md,Node)))
        assert(not(isinstance(self.md_root,Node)))

        # confirm objects have inherited Node attrs
        assert(hasattr(self.ar,"_branch"))
        assert(hasattr(self.ar,"_metadata"))
        assert(hasattr(self.pl,"_branch"))
        assert(hasattr(self.pl,"_metadata"))
        assert(hasattr(self.pla,"_branch"))
        assert(hasattr(self.pla,"_metadata"))
        assert(not(hasattr(self.md,"_branch")))
        assert(not(hasattr(self.md,"_metadata")))


    def test_structure(self):
        """
        Confirm that the structure of the tree is correct, and
        that objects can be retrieved as expected
        """
        # confirm that the data is where it should be
        # and is accessible via relative paths
        assert(self.root.tree('array') == self.ar)
        assert(self.root.tree('array/pointlistarray') == self.pla)
        assert(self.root.tree('array/pointlistarray/pointlist') == self.pl)
        assert(self.root.tree('pointlist3/pointlist4').metadata['metadata3'] == \
            self.md3)
        assert(self.pla.tree('pointlist2') == self.pl2)
        assert(self.ar.metadata['metadata'] == self.md)
        assert(self.ar.metadata['metadata2'] == self.md2)
        assert(self.root.tree('pointlist3') == self.pl3)
        assert(self.root.metadata['metadata_root'] == self.md_root)
        assert(self.root.metadata['metadata_root2'] == self.md_root2)

        # confirm data is accessible via root paths
        assert(self.ar.tree('/array') == self.ar)
        assert(self.pl4.tree('/array') == self.ar)
        assert(self.pla.tree('/array/pointlistarray/pointlist') == self.pl)

        # confirm correct ._treepath attrs
        assert(self.ar._treepath == '/array')
        assert(self.pla._treepath == '/array/pointlistarray')
        assert(self.pl3._treepath == '/pointlist3')

        # confirm that all nodes point to the same root node
        assert(self.root == self.pla.root == self.pl.root == self.pl2.root == \
            self.pl3.root == self.pl4.root)


    def test_branchcut(self):
        """
        Confirms correct branch cutting behavior
        for default metadata behavior (transfer root metadata)
        """
        # cutting off a branch
        new_root = self.pl4.tree(cut=True)
        assert(self.pl4.root == new_root)
        assert(new_root.tree('pointlist4') == self.pl4)
        assert(new_root.metadata == self.root.metadata)

    def test_branchcut_no_metadata(self):
        """
        Confirms correct branch cutting behavior
        without transferring old root metadata
        """
        # cutting off a branch
        new_root = self.pl4.tree(cut=False)
        assert(len(new_root.metadata)==0)

    def test_branchcut_copy_metadata(self):
        """
        Confirms correct branch cutting behavior
        when copying root metadata
        """
        # cutting off a branch
        new_root = self.pl4.tree(cut='copy')
        assert(len(new_root.metadata)==2)
        assert(new_root.metadata != self.root.metadata)

    def test_graft(self):
        """
        Confirms correct grafting behavior
        """
        # make a new tree to graft onto
        new_root = self.pl4.tree(cut=True)
        # graft
        self.pla.tree( graft=new_root.tree('pointlist4') )
        # check
        data = self.root.tree('array/pointlistarray/pointlist4')
        assert(isinstance(data,PointList))
        assert(data == self.pl4)

    def test_graft2(self):
        """
        Confirms correct grafting behavior with alternative syntax
        """
        # make a new tree to graft onto
        new_root = self.pl4.tree(cut=True)
        # graft
        self.pla.graft( new_root.tree('pointlist4') )
        # check
        data = self.root.tree('array/pointlistarray/pointlist4')
        assert(isinstance(data,PointList))
        assert(data == self.pl4)



    def test_graft3(self):
        """
        Confirms correct grafting behavior with alternative syntax
        """
        # make a new tree to graft onto
        new_root = self.pl4.tree(cut=True)
        # graft
        self.pla.tree( graft=(new_root.tree('pointlist4'),True) )
        # check
        data = self.root.tree('array/pointlistarray/pointlist4')
        assert(isinstance(data,PointList))
        assert(data == self.pl4)




class TestTreeWrite(TreeBuilder):

    def setup_cls(cls):
        cls._clear_files()

    def teardown_cls(cls):
        cls._clear_files()

    def setup_method(self):
        self._build_tree()

    def teardown_method(self):
        self._del_tree()

    def teardown_cls(cls):
        cls._clear_files()

    def test_write_single_node(self):
        save(
            single_node_path,
            self.ar,
            tree = False,
            mode = 'o'
        )

    def test_write_unrooted_node(self):
        save(
            unrooted_node_path,
            self.ar_unrooted,
            mode = 'o'
        )
        pass

    def test_write_whole_tree(self):
        print('Tree:')
        self.root.tree(True)
        print()
        save(
            whole_tree_path,
            self.root,
            tree = True,
            mode = 'o'
        )
        pass

    def test_write_h5_subtree(self):
        """
        Write a subtree to file
        """
        save(
            subtree_path,
            self.pl3,
            tree = True,
            mode = 'o'
        )

    def test_write_h5_subtree_noroot(self):
        """
        Write a subtree tree to file without it's root dataset
        """
        save(
            subtree_noroot_path,
            self.pl3,
            tree = None,
            mode = 'o'
        )

    def test_append_to_h5(self):
        """
        Append to an existing h5 file
        """
        save(
            subtree_append_to_path,
            self.ar,
            tree = False,
            mode = 'o'
        )
        save(
            subtree_append_to_path,
            self.ar2,
            tree = True,
            mode = 'a'
        )
        pass

    def test_append_over_h5(self):
        """
        Append to an existing h5 file, overwriting a redundantly named object
        """
        save(
            subtree_append_over_path,
            self.ar,
            tree = False,
            mode = 'o'
        )
        save(
            subtree_append_over_path,
            self.ar2.root,
            tree = True,
            mode = 'ao'
        )
        pass

    def _clear_files(self):
        """
        Delete h5 files which this test suite wrote
        """
        paths = [
            single_node_path,
            unrooted_node_path,
            whole_tree_path,
            subtree_path,
            subtree_noroot_path,
            subtree_append_to_path,
            subtree_append_over_path
        ]
        for p in paths:
            if exists(p):
                remove(p)














class TestTreeRead(TreeBuilder):

    @classmethod
    def setup_class(cls):
        cls._clear_files(cls)
        cls._build_tree(cls)
        cls._write_h5_single_node(cls)
        cls._write_h5_unrooted_node(cls)
        cls._write_h5_whole_tree(cls)
        cls._write_h5_subtree(cls)
        cls._write_h5_subtree_noroot(cls)
        cls._append_to_h5(cls)
        cls._append_over_h5(cls)

    @classmethod
    def teardown_class(cls):
        cls._clear_files(cls)

    def setup_method(self, method):
        self._build_tree()

    def teardown_method(self, method):
        self._del_tree()


    @staticmethod
    def pointlist_equal(pl1,pl2):
        """
        Return true iff pl1 and pl2 are equal
        """
        # allow for pl1 and pl2 to have the same fields in different order
        k1,k2 = pl1.fields,pl2.fields
        assert(all([k in k1 for k in k2]))
        assert(all([k in k2 for k in k1]))
        for k in k1:
            try:
                assert(array_equal(pl1.data[k],pl2.data[k]))
            except AssertionError:
                return False
        return True


    # Tests

    def test_file_validity(self):
        assert(_is_EMD_file(single_node_path))
        assert(_is_EMD_file(unrooted_node_path))
        assert(_is_EMD_file(whole_tree_path))
        assert(_is_EMD_file(subtree_path))
        assert(_is_EMD_file(subtree_noroot_path))
        assert(_is_EMD_file(subtree_append_to_path))
        assert(_is_EMD_file(subtree_append_over_path))


    def test_hdf5_group_paths(self):
        """
        Confirms that the _treepath is always the same as
        the group path without its root
        """
        # Open the file
        with h5py.File(whole_tree_path,'r') as f:

            # loop over nodes
            for node in (
                self.root,
                self.ar,
                self.pl,
                self.pl2,
                self.pl3,
                self.pl4,
                self.pla,
            ):
                assert( f[f'/{self.root.name}/' + node._treepath] )


    def test_single_node_io(self):
        """ Should contain a single array 'array' inside a single root 'root',
        and 'array' should match self.ar
        """
        # Load data
        loaded_data = read(
            single_node_path
        )

        # Does the tree look as it should?
        assert(isinstance(loaded_data,Array))
        lar = loaded_data

        # Did the root metadata load correctly
        lmd1 = lar.root.metadata['metadata_root']
        lmd2 = lar.root.metadata['metadata_root2']
        assert(lmd1 is not self.md_root)
        assert(lmd2 is not self.md_root2)
        assert(lmd1._params == self.md_root._params)
        assert(lmd2._params == self.md_root2._params)

        # Did the array load correctly?
        assert(lar is not self.ar)
        assert(lar.name == self.ar.name)
        assert(lar.units == self.ar.units)
        assert(array_equal(lar.data,self.ar.data))
        for i in range(len(lar.dims)):
            assert(array_equal(lar.dims[i],self.ar.dims[i]))
            assert(lar.dim_names[i] == self.ar.dim_names[i])
            assert(lar.dim_units[i] == self.ar.dim_units[i])
        pass



    def test_single_node_io_emdpath(self):
        """ Should contain a single array 'array' inside a single root 'root',
        and 'array' should match self.ar
        """
        # Load data
        lar = read(
            single_node_path,
            emdpath = '/root/array'
        )

        # Does the tree look as it should?
        assert(isinstance(lar,Array))
        assert(isinstance(lar.root,Root))

        # Did the root metadata load correctly
        lmd1 = lar.root.metadata['metadata_root']
        lmd2 = lar.root.metadata['metadata_root2']
        assert(lmd1 is not self.md_root)
        assert(lmd2 is not self.md_root2)
        assert(lmd1._params == self.md_root._params)
        assert(lmd2._params == self.md_root2._params)

        # Did the array load correctly?
        assert(lar is not self.ar)
        assert(lar.name == self.ar.name)
        assert(lar.units == self.ar.units)
        assert(array_equal(lar.data,self.ar.data))
        for i in range(len(lar.dims)):
            assert(array_equal(lar.dims[i],self.ar.dims[i]))
            assert(lar.dim_names[i] == self.ar.dim_names[i])
            assert(lar.dim_units[i] == self.ar.dim_units[i])
        pass


    def test_unrooted_node_io(self):
        """ Should contain a single array 'array' inside a single root 'root',
        and 'array' should match self.ar
        """
        # Load data
        loaded_data = read(
            unrooted_node_path
        )

        # Does the tree look as it should?
        assert(isinstance(loaded_data,Array))
        assert(loaded_data.name == 'unrooted_array')
        lar = loaded_data

        # Did the root metadata load correctly
        assert(len(lar.root.metadata.keys())==0)

        # Did the array load correctly?
        assert(lar is not self.ar_unrooted)
        assert(lar.name == self.ar_unrooted.name)
        assert(lar.units == self.ar_unrooted.units)
        assert(array_equal(lar.data,self.ar_unrooted.data))
        for i in range(len(lar.dims)):
            assert(array_equal(lar.dims[i],self.ar_unrooted.dims[i]))
            assert(lar.dim_names[i] == self.ar_unrooted.dim_names[i])
            assert(lar.dim_units[i] == self.ar_unrooted.dim_units[i])

        # Did the dim names/units pass through correctly?
        assert(lar.get_dim_name(0) == 'pony_axis')
        assert(lar.get_dim_units(0) == 'ponies')
        assert(lar.get_dim_name(1) == 'dim1')
        assert(lar.get_dim_units(1) == 'cows')



    def test_whole_tree_io(self):
        """ Should contain the full data tree:

            Root                  'root'
              |-Metadata          'metadata_root'
              |-Metadata          'metadata_root'
              |-PointList         'pointlist3'
              | |-PointList       'pointlist4'
              |   |-Metadata      'metadata3'
              |-Array             'array
                |-Metadata        'metadata'
                |-Metadata        'metadata2'
                |-PointListArray  'pointlistarray'
                  |-PointList     'pointlist'
                  |-PointList     'pointlist2'
        """
        # Load data
        loaded_data = read(
            whole_tree_path
        )

        # Does the tree look as it should?
        assert(isinstance(loaded_data,Root))
        assert(loaded_data.name == 'root')

        lmd_root = loaded_data.metadata['metadata_root']
        lmd_root2 = loaded_data.metadata['metadata_root2']
        lpl3 = loaded_data.tree('/pointlist3')
        lpl4 = loaded_data.tree('pointlist3/pointlist4')
        lmd3 = lpl4.metadata['metadata3']
        lar = loaded_data.tree('array')
        lmd = lar.metadata['metadata']
        lmd2 = lar.metadata['metadata2']
        lpla = loaded_data.tree('array/pointlistarray')
        lpl = loaded_data.tree('array/pointlistarray/pointlist')
        lpl2 = loaded_data.tree('array/pointlistarray/pointlist2')

        # Check types
        assert(isinstance(lmd_root,Metadata))
        assert(isinstance(lmd_root2,Metadata))
        assert(isinstance(lpl3,PointList))
        assert(isinstance(lpl4,PointList))
        assert(isinstance(lmd3,Metadata))
        assert(isinstance(lar,Array))
        assert(isinstance(lmd,Metadata))
        assert(isinstance(lmd2,Metadata))
        assert(isinstance(lpla,PointListArray))
        assert(isinstance(lpl,PointList))
        assert(isinstance(lpl2,PointList))

        # New objects are distinct from old objects
        assert(lmd_root is not self.md_root)
        assert(lmd_root2 is not self.md_root2)
        assert(lpl3 is not self.pl3)
        assert(lpl4 is not self.pl4)
        assert(lmd3 is not self.md3)
        assert(lar is not self.ar)
        assert(lmd is not self.md)
        assert(lmd2 is not self.md2)
        assert(lpla is not self.pla)
        assert(lpl is not self.pl)
        assert(lpl2 is not self.pl2)

        # New objects carry identical data to old objects
        assert(lmd_root._params == self.md_root._params)
        assert(lmd_root2._params == self.md_root2._params)
        assert(self.pointlist_equal(lpl3, self.pl3))
        assert(self.pointlist_equal(lpl4, self.pl4))
        assert(lmd3._params == self.md3._params)
        assert(array_equal(lar.data, self.ar.data))
        assert(lmd._params == self.md._params)
        assert(lmd2._params == self.md2._params)
        assert(self.pointlist_equal(lpl, self.pl))
        assert(self.pointlist_equal(lpl2, self.pl2))
        for x in range(lpla.shape[0]):
            for y in range(lpla.shape[1]):
                assert(self.pointlist_equal(lpla[x,y], self.pla[x,y]))

        pass


    def test_whole_tree_io_treeIsFalse(self):
        """ Should contain only:

            Root                  'root'
              |-PointListArray    'pointlistarray'
        """
        # Load data
        loaded_data = read(
            whole_tree_path,
            emdpath = '/root/array/pointlistarray',
            tree = False
        )

        # Does the tree look as it should?
        assert(isinstance(loaded_data,PointListArray))
        assert(loaded_data.name == 'pointlistarray')

        lmd_root = loaded_data.root.metadata['metadata_root']
        lmd_root2 = loaded_data.root.metadata['metadata_root2']

        # Check types
        assert(isinstance(lmd_root,Metadata))
        assert(isinstance(lmd_root2,Metadata))

        # New objects are distinct from old objects
        assert(lmd_root is not self.md_root)
        assert(lmd_root2 is not self.md_root2)
        assert(loaded_data is not self.pla)

        # New objects carry identical data to old objects
        assert(lmd_root._params == self.md_root._params)
        assert(lmd_root2._params == self.md_root2._params)
        for x in range(loaded_data.shape[0]):
            for y in range(loaded_data.shape[1]):
                assert(self.pointlist_equal(loaded_data[x,y], self.pla[x,y]))

        pass


    def test_subtree_io(self):
        """ Tree should be root/pointlist3/pointlist4, plus
        metadata at root and pl4
        """
        # Load data
        loaded_data = read(
            subtree_path
        )

        # Does the tree look as it should?
        assert(isinstance(loaded_data,PointList))
        assert(loaded_data.name == 'pointlist3')

        lmd_root = loaded_data.root.metadata['metadata_root']
        lmd_root2 = loaded_data.root.metadata['metadata_root2']
        lpl3 = loaded_data
        lpl4 = loaded_data.tree('pointlist4')
        lmd3 = lpl4.metadata['metadata3']

        # Check types
        assert(isinstance(lmd_root,Metadata))
        assert(isinstance(lmd_root2,Metadata))
        assert(isinstance(lpl3,PointList))
        assert(isinstance(lpl4,PointList))
        assert(isinstance(lmd3,Metadata))

        # New objects are distinct from old objects
        assert(lmd_root is not self.md_root)
        assert(lmd_root2 is not self.md_root2)
        assert(lpl3 is not self.pl3)
        assert(lpl4 is not self.pl4)
        assert(lmd3 is not self.md3)

        # New objects carry identical data to old objects
        assert(lmd_root._params == self.md_root._params)
        assert(lmd_root2._params == self.md_root2._params)
        assert(self.pointlist_equal(lpl3, self.pl3))
        assert(self.pointlist_equal(lpl4, self.pl4))
        assert(lmd3._params == self.md3._params)

        pass



    def test_subtree_noroot_io(self):
        """ Tree should be root/pointlist4, plus
        metadata at root and pl4
        """
        # Load data
        loaded_data = read(
            subtree_noroot_path
        )

        # Does the tree look as it should?
        assert(isinstance(loaded_data,PointList))
        assert(loaded_data.name == 'pointlist4')

        lmd_root = loaded_data.root.metadata['metadata_root']
        lmd_root2 = loaded_data.root.metadata['metadata_root2']
        lpl4 = loaded_data
        lmd3 = lpl4.metadata['metadata3']

        # Check types
        assert(isinstance(lmd_root,Metadata))
        assert(isinstance(lmd_root2,Metadata))
        assert(isinstance(lmd3,Metadata))

        # New objects are distinct from old objects
        assert(lmd_root is not self.md_root)
        assert(lmd_root2 is not self.md_root2)
        assert(lpl4 is not self.pl4)
        assert(lmd3 is not self.md3)

        # New objects carry identical data to old objects
        assert(lmd_root._params == self.md_root._params)
        assert(lmd_root2._params == self.md_root2._params)
        assert(self.pointlist_equal(lpl4, self.pl4))
        assert(lmd3._params == self.md3._params)

        pass


    def test_append_to_io(self):
        """ Tree should be
                root
                    |--array
                        |--pointlist5
        """
        # Load data
        loaded_data = read(
            subtree_append_to_path
        )

        # Does the tree look as it should?
        assert(isinstance(loaded_data,Array))
        assert(loaded_data.name == 'array')

        lmd_root = loaded_data.root.metadata['metadata_root']
        lmd_root2 = loaded_data.root.metadata['metadata_root2']
        lar = loaded_data
        lmd = lar.metadata['metadata']
        lmd2 = lar.metadata['metadata2']
        lpl5 = loaded_data.tree('pointlist5')

        # Ensure the data that shouldn't be here, isn't
        assert('pointlist3' not in loaded_data.root._branch.keys())
        assert('array' in loaded_data.root._branch.keys())

        # Check types
        assert(isinstance(lmd_root,Metadata))
        assert(isinstance(lmd_root2,Metadata))
        assert(isinstance(lar,Array))
        assert(isinstance(lmd,Metadata))
        assert(isinstance(lmd2,Metadata))
        assert(isinstance(lpl5,PointList))

        # New objects are distinct from old objects
        assert(lmd_root is not self.md_root)
        assert(lmd_root2 is not self.md_root2)
        assert(lar is not self.ar)
        assert(lmd is not self.md)
        assert(lmd2 is not self.md2)
        assert(lpl5 is not self.pl5)

        # New objects carry identical data to old objects
        assert(lmd_root._params == self.md_root._params)
        assert(lmd_root2._params == self.md_root2._params)
        assert(array_equal(lar.data, self.ar.data))
        assert(not(array_equal(lar.data, self.ar2.data))) # ensure we didn't overwrite
        assert(lmd._params == self.md._params)
        assert(lmd2._params == self.md2._params)
        assert(self.pointlist_equal(lpl5, self.pl5))

        pass


    def test_append_over_io(self):
        """ Tree should be
                root
                    |--array
                        |--pointlist5
        where array should have no metadata and its data should
        be that of ar2 and not ar1
        """
        # Load data
        loaded_data = read(
            subtree_append_over_path
        )

        # Does the tree look as it should?
        assert(isinstance(loaded_data,Array))
        assert(loaded_data.name == 'array')

        lmd_root = loaded_data.root.metadata['metadata_root']
        lmd_root2 = loaded_data.root.metadata['metadata_root2']
        assert(len(loaded_data.metadata)==0)
        lpl5 = loaded_data.tree('pointlist5')

        # Ensure the data that shouldn't be here, isn't
        assert('pointlist3' not in loaded_data.root._branch.keys())
        assert('array' in loaded_data.root._branch.keys())

        # Check types
        assert(isinstance(lmd_root,Metadata))
        assert(isinstance(lmd_root2,Metadata))
        assert(isinstance(lpl5,PointList))

        # New objects are distinct from old objects
        assert(lmd_root is not self.md_root)
        assert(lmd_root2 is not self.md_root2)
        assert(loaded_data is not self.ar2)
        assert(lpl5 is not self.pl5)

        # New objects carry identical data to old objects
        assert(lmd_root._params == self.md_root._params)
        assert(lmd_root2._params == self.md_root2._params)
        assert(array_equal(loaded_data.data, self.ar2.data))    # ensure we did overwrite
        assert(not(array_equal(loaded_data.data, self.ar.data)))
        assert(self.pointlist_equal(lpl5, self.pl5))

        pass






    # setup/teardown utilities

    def _write_h5_single_node(self):
        """
        Write a single node to file
        """
        save(
            single_node_path,
            self.ar,
            tree = False
        )

    def _write_h5_unrooted_node(self):
        """
        Write an unrooted node to file
        """
        save(
            unrooted_node_path,
            self.ar_unrooted
        )

    def _write_h5_whole_tree(self):
        """
        Write the whole tree to file
        """
        save(
            whole_tree_path,
            self.root,
            tree = True
        )

    def _write_h5_subtree(self):
        """
        Write a subtree to file
        """
        save(
            subtree_path,
            self.pl3,
            tree = True
        )

    def _write_h5_subtree_noroot(self):
        """
        Write a subtree tree to file without it's root dataset
        """
        save(
            subtree_noroot_path,
            self.pl3,
            tree = None
        )

    def _append_to_h5(self):
        """
        Append to an existing h5 file
        """
        save(
            subtree_append_to_path,
            self.ar,
            tree = False
        )
        save(
            subtree_append_to_path,
            self.ar2,
            tree = True,
            mode = 'a'
        )
        pass

    def _append_over_h5(self):
        """
        Append to an existing h5 file, overwriting a redundantly named object
        """
        save(
            subtree_append_over_path,
            self.ar,
            tree = False
        )
        save(
            subtree_append_over_path,
            self.ar2.root,
            tree = True,
            mode = 'ao'
        )
        pass

    def _clear_files(self):
        """
        Delete h5 files which this test suite wrote
        """
        paths = [
            single_node_path,
            unrooted_node_path,
            whole_tree_path,
            subtree_path,
            subtree_noroot_path,
            subtree_append_to_path,
            subtree_append_over_path
        ]
        for p in paths:
            if exists(p):
                remove(p)









from emdfile import Metadata, Node, Root, save, read, _TESTPATH

import numpy as np
from os.path import join, exists
from os import remove

path = join(_TESTPATH, 'test_metadata.h5')



class TestMetadata:

    # setup + teardown

    def setup_method(self):
        """
        Make a simple and a nested Metadata instance.
        Clear filepaths.
        """
        self._clear_data()

        m = Metadata( name='test_metadata' )

        m['x'] = 10
        m['y'] = True
        m['z'] = None
        m['a'] = (1,2,3)
        m['b'] = np.ones((4,5),dtype='uint16')
        m['c'] = (np.ones(5,dtype=bool), np.ones(6,dtype=np.float32))

        n = Metadata( name='test_nested_metadata' )

        n['x'] = 10
        n['y'] = {'arg':1,'barg':(1,2,3),'targ':None}
        n['z'] = {}
        n['z']['a'] = 'moop'
        n['z']['b'] = {}
        n['z']['b']['c'] = True

        self.m = m
        self.n = n
        pass

    def teardown_method(self):
        """
        Clear filepaths.
        """
        self._clear_data()

    def _clear_data(self):
        if exists(path):
            remove(path)
            pass



    # tests

    def test_metadata_basic_io(self):
        """
        Make a metadata instance
        store a piece of data
        save, read
        """
        m = self.m
        save(path, m)
        _m = read(path)

        assert(m['x'] == _m['x'])
        assert(m['y'] == _m['y'])
        assert(m['z'] == _m['z'])
        assert(m['a'] == _m['a'])
        assert(np.array_equal(m['b'], _m['b']))
        assert(np.array_equal(m['c'][0], _m['c'][0]))
        assert(np.array_equal(m['c'][1], _m['c'][1]))
        assert(m['b'].dtype == _m['b'].dtype)
        assert(m['c'][0].dtype == _m['c'][0].dtype)
        assert(m['c'][1].dtype == _m['c'][1].dtype)
        pass


    def test_metadata_nested_io(self):
        """
        Make a metadata instance
        store some nested metadata
        save, read
        """
        n = self.n
        save(path, n)
        _n = read(path)

        assert(n['x'] == _n['x'])
        assert(n['y'] == _n['y'])
        assert(n['z'] == _n['z'])
        pass


    def test_empty_lists_and_tups(self):
        """
        """
        m = Metadata( name='metadata' )
        m['x'] = ()
        m['y'] = []
        save(path, m)
        n = read(path)
        assert(n['x'] == ())
        assert(n['y'] == [])









# classes

from emdfile.classes import (
    Metadata,
    Node,
    Root,
    Array,
    PointList,
    PointListArray,
    Custom
)





# header hooks

_PROGRAM_NAME = 'emdfile'
_USER_NAME = ''
def set_author(author):
    """ Accepts a string, which will be written to the "authoring_user" field in any EMD file headers
        written during this Python session
    """
    global _USER_NAME
    _USER_NAME = author





# read/write

from emdfile.read import (
    read,
    print_h5_tree,
    _is_EMD_file,
    _get_EMD_version,
    _version_is_geq,
    _read_metadata
)
from emdfile.write import write as save




# version

from emdfile.version import __version__





# n-dimensional progress bar

from emdfile.tqdmnd import tqdmnd




# test paths

from os.path import dirname,join
_TESTPATH = join(join(dirname(__file__), "../.."), "test/unit_test_data")







import numpy as np
from os.path import join,exists
from os import remove
from numpy import array_equal

from emdfile import _TESTPATH
from emdfile import save,read
from emdfile.read import _is_EMD_file,_get_EMD_rootgroups
from emdfile.classes import (
    Node,
    Root,
    Metadata,
    Array,
    PointList,
    PointListArray
)

# Set paths
dirpath = _TESTPATH
path_h5 = join(dirpath,"test.h5")





class TestEmd:

    ## Setup and teardown

    @classmethod
    def setup_class(cls):
        cls._clear_files(cls)
        cls._make_data(cls)

    @classmethod
    def teardown_class(cls):
        pass

    def setup_method(self, method):
        pass

    def teardown_method(self, method):
        self._clear_files()

    def _make_data(self):
        """ Make
            - an array
            - an array with some 'slicelabels'
            #- a pointlist
            #- a pointlistarray
        """
        # arrays
        self.array = Array(
            data = np.arange(np.prod((4,8))).reshape((4,8))
        )
        self.array2 = Array(
            data = np.arange(np.prod((4,8,2))).reshape((4,8,2)),
            name = 'array2',
            units = 'meowths',
            dims = [
                [0,2],
                [0,2]
            ],
            dim_units = [
                'earthmiles',
                'pokemonmiles'
            ],
            dim_names = [
                'pokemonblue',
                'pokemonred'
            ],
            slicelabels = [
                'a',
                'b'
            ]
        )

        # pointlist
        # pointlistarray





    def _clear_files(self):
        """
        Delete h5 files which this test suite wrote
        """
        paths = [
            path_h5
        ]
        for p in paths:
            if exists(p):
                remove(p)





    ## Tests

    def test_array(self):
        """ Save then read as array, and compare its contents before/after
        """
        assert(isinstance(self.array,Array))
        # save and read
        print(path_h5)
        save(path_h5,self.array)
        new_array = read(path_h5)
        # check it's the same
        assert(isinstance(new_array,Array))
        assert(array_equal(self.array.data,new_array.data))

    def test_array2(self):
        """ Save then read as array, and compare its contents before/after
        """
        assert(isinstance(self.array2,Array))
        # save and read
        save(path_h5,self.array2)
        new_array = read(path_h5)
        # check it's the same
        assert(isinstance(new_array,Array))
        assert(array_equal(self.array2.data,new_array.data))
        # check that metadata passed through
        assert(self.array2.name == new_array.name)
        assert(self.array2.units == new_array.units)
        assert(array_equal(self.array2.dims[0],new_array.dims[0]))
        assert(array_equal(self.array2.dims[1],new_array.dims[1]))
        for i in range(2):
            assert(self.array2.dim_units[i] == new_array.dim_units[i])
            assert(self.array2.dim_names[i] == new_array.dim_names[i])
            assert(self.array2.slicelabels[i] == new_array.slicelabels[i])

